#
#----------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#----------------------------------------------------------------------

from .uvm_phase import UVMPhase
from .uvm_object_globals import (UVM_DEBUG, UVM_PHASE_ENDED, UVM_PHASE_EXECUTING, UVM_PHASE_IMP,
                                 UVM_PHASE_READY_TO_END, UVM_PHASE_STARTED)
from .uvm_domain import UVMDomain
from .uvm_debug import uvm_debug
from .uvm_globals import uvm_report_fatal, uvm_report_info


class UVMBottomupPhase(UVMPhase):
    """
    Base class for function phases that operate bottom-up.
    The pure virtual function execute() is called for each component.
    This is the default traversal so is included only for naming.

    A bottom-up function phase completes when the <execute()> method
    has been called and returned on all applicable components
    in the hierarchy.
    """


    def __init__(self, name):
        """
        Create a new instance of a bottom-up phase.

        Args:
            name (str): Name of the phase.
        """
        UVMPhase.__init__(self, name, UVM_PHASE_IMP)


    def traverse(self, comp, phase, state):
        """
        Traverses the component tree in bottom-up order, calling `execute` for
        each component.

        Args:
            comp (UVMComponent): Top-level component for traversal
            phase (UVMPhase):
            state:
        """
        uvm_debug(self, 'traverse', self.get_name() + ' traversing bottomup phase now with ' +
                comp.get_name())
        name = ""
        phase_domain = phase.get_domain()
        comp_domain = comp.get_domain()

        if comp.has_first_child():
            child = comp.get_first_child()
            while child is not None:
                self.traverse(child, phase, state)
                child = comp.get_next_child()

        if UVMPhase.m_phase_trace:
            dom_name = "unknown"
            if comp_domain is not None:
                dom_name = comp_domain.get_name()
            uvm_report_info("PH_TRACE",("bottomup-phase phase={} state={} comp={} comp.domain={} phase.domain={}"
                  .format(phase.get_name(), str(state), comp.get_full_name(),
                      dom_name, phase_domain.get_name())), UVM_DEBUG)

        if phase_domain == UVMDomain.get_common_domain() or phase_domain == comp_domain:
            if state == UVM_PHASE_STARTED:
                comp.m_current_phase = phase
                comp.m_apply_verbosity_settings(phase)
                comp.phase_started(phase)
            elif state == UVM_PHASE_EXECUTING:
                ph = self
                if self in comp.m_phase_imps:
                    ph = comp.m_phase_imps[self]
                ph.execute(comp, phase)
            elif state == UVM_PHASE_READY_TO_END:
                comp.phase_ready_to_end(phase)
            elif state == UVM_PHASE_ENDED:
                comp.phase_ended(phase)
                comp.m_current_phase = None
            else:
                uvm_report_fatal("PH_BADEXEC", "bottomup phase traverse internal error")

    def execute(self, comp, phase):
        """
        Executes the bottom-up phase `phase` for the component `comp`.

        Args:
            comp (UVMComponent):
            phase (UVMPhase):
        """
        # reseed this process for random stability
        #process proc = process::self()
        #proc.srandom(uvm_create_random_seed(phase.get_type_name(), comp.get_full_name()))
        comp.m_current_phase = phase
        self.exec_func(comp,phase)



#------------------------------------------------------------------------------
#     Copyright 2007-2011 Mentor Graphics Corporation
#     Copyright 2007-2011 Cadence Design Systems, Inc.
#     Copyright 2010 Synopsys, Inc.
#     Copyright 2013 Cisco Systems, Inc.
#     Copyright 2019 Tuomas Poikela
#     All Rights Reserved Worldwide
#
#     Licensed under the Apache License, Version 2.0 (the "License"); you may not
#     use this file except in compliance with the License.    You may obtain a copy
#     of the License at
#
#             http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#     WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    See the
#     License for the specific language governing permissions and limitations
#     under the License.
#------------------------------------------------------------------------------

from .uvm_component import UVMComponent
from .uvm_object_globals import (UVM_EXPORT, UVM_IMPLEMENTATION, UVM_NONE, UVM_PHASE_DONE,
                                 UVM_PHASE_EXECUTING, UVM_PORT)
from .uvm_domain import end_of_elaboration_ph
from .sv import sv
from ..macros.uvm_message_defines import uvm_report_fatal

UVM_UNBOUNDED_CONNECTIONS = -1
s_connection_error_id = "Connection Error"
s_connection_warning_id = "Connection Warning"
s_spaces = "                                             "

ERR_MSG2 = ("Port parent: {}. Cannot call an imp port's connect method. "
    + "An imp is connected only to the component passed in its constructor."
    + "(You attempted to bind this imp to {})")

ERR_MSG3 = ("Cannot connect exports to ports Try calling port.connect(export)"
    + " instead. (You attempted to bind this export to {}).")

## typedef uvm_port_component_base uvm_port_list[string]


class UVMPortComponentBase(UVMComponent):
    """
    This class defines an interface for obtaining a port's connectivity lists
    after or during the end_of_elaboration phase.    The sub-class,
    `UVMPortComponent` implements this interface.

    The connectivity lists are returned in the form of handles to objects of this
    type. This allowing traversal of any port's fan-out and fan-in network
    through recursive calls to <get_connected_to> and <get_provided_to>. Each
    port's full name and type name can be retrieved using ~get_full_name~ and
    ~get_type_name~ methods inherited from <uvm_component>.
    """


    def __init__(self, name, parent):
        UVMComponent.__init__(self, name, parent)

    # Function: get_connected_to
    #
    # For a port or export type, this function fills ~list~ with all
    # of the ports, exports and implementations that this port is
    # connected to.
    def get_connected_to(self, port_list):
        pass  # pure virtual

    # Function: get_provided_to
    #
    # For an implementation or export type, this function fills ~list~ with all
    # of the ports, exports and implementations that this port is
    # provides its implementation to.
    def get_provided_to(self, port_list):
        pass  # pure virtual

    def is_port(self):
        """
        Function: is_port

        """
        pass  # pure virtual

    def is_export(self):
        """
        Function: is_export

        """
        pass  # pure virtual

    # Function: is_imp
    #
    # These function determine the type of port. The functions are
    # mutually exclusive; one will return 1 and the other two will
    # return 0.
    def is_imp(self):
        pass  # pure virtual

    def build_phase(self, phase):
        """
        Turn off auto config by not calling build_phase()

        Args:
            phase:
        """
        self.build()  # for backward compat
        return

    def do_task_phase(self, phase):
        pass



class UVMPortComponent(UVMPortComponentBase):
    """
    See description of `UVMPortComponentBase` for information about this class
    """

    def __init__(self, name, parent, port):
        UVMPortComponentBase.__init__(self, name, parent)
        if port is None:
            uvm_report_fatal("Bad usage", "Null handle to port", UVM_NONE)
        self.m_port = port

    def get_type_name(self):
        if self.m_port is None:
            return "uvm_port_component"
        return self.m_port.get_type_name()

    def resolve_bindings(self):
        self.m_port.resolve_bindings()

    def get_port(self):
        """
        Function: get_port

        Retrieve the actual port object that this proxy refers to.
        Returns:
        """
        return self.m_port

    def get_connected_to(self, port_list):
        self.m_port.get_connected_to(port_list)

    def get_provided_to(self, port_list):
        self.m_port.get_provided_to(port_list)

    def is_port(self):
        return self.m_port.is_port()

    def is_export(self):
        return self.m_port.is_export()

    def is_imp(self):
        return self.m_port.is_imp()

## virtual class uvm_port_base #(type IF=uvm_void) extends IF


class UVMPortBase():
    """
    Transaction-level communication between components is handled via its ports,
    exports, and imps, all of which derive from this class.

    The UVM provides a complete set of ports, exports, and imps for the OSCI-
    standard TLM interfaces. They can be found in the ../src/tlm/ directory.

    Just before `UVMComponent.end_of_elaboration_phase`, an internal
    `UVMComponent.resolve_bindings` process occurs, after which each port and
    export holds a list of all imps connected to it via hierarchical connections
    to other ports and exports. In effect, we are collapsing the port's fanout,
    which can span several levels up and down the component hierarchy, into a
    single array held local to the port. Once the list is determined, the port's
    min and max connection settings can be checked and enforced.

    uvm_port_base possesses the properties of components in that they have a
    hierarchical instance path and parent. Because SystemVerilog does not support
    multiple inheritance, uvm_port_base cannot extend both the interface it
    implements and <uvm_component>. Thus, uvm_port_base contains a local instance
    of uvm_component, to which it delegates such commands as get_name,
    get_full_name, and get_parent.
    """




    # Function: new
    #
    # The first two arguments are the normal <uvm_component> constructor
    # arguments.
    #
    # The ~port_type~ can be one of <UVM_PORT>, <UVM_EXPORT>, or
    # <UVM_IMPLEMENTATION>.
    #
    # The ~min_size~ and ~max_size~ specify the minimum and maximum number of
    # implementation (imp) ports that must be connected to this port base by the
    # end of elaboration. Setting ~max_size~ to ~UVM_UNBOUNDED_CONNECTIONS~ sets no
    # maximum, i.e., an unlimited number of connections are allowed.
    #
    # By default, the parent/child relationship of any port being connected to
    # this port is not checked. This can be overridden by configuring the
    # port's ~check_connection_relationships~ bit via ~uvm_config_int::set()~. See
    # <connect> for more information.
    def __init__(self, name, parent, port_type, min_size=0, max_size=1):
        self.m_port_type = port_type
        self.m_min_size    = min_size
        self.m_max_size    = max_size
        self.m_comp = UVMPortComponent(name, parent, self)
        self.m_resolved = False
        self.m_imp_list = {}
        self.m_if_mask = 0
        self.m_def_index = 0
        self.m_provided_by = {}
        self.m_provided_to = {}
        self.m_if = None
        self.m_imp = None
        #if (!uvm_config_int::get(m_comp, "", "check_connection_relationships",tmp))
        #   self.m_comp.set_report_id_action(s_connection_warning_id, UVM_NO_ACTION)

    def get_name(self):
        """ Returns the leaf name of this port. """
        return self.m_comp.get_name()

    def get_full_name(self):
        """ Returns the full hierarchical name of this port. """
        return self.m_comp.get_full_name()


    def get_parent(self):
        """ Returns the handle to this port's parent, or ~null~ if it has no
        parent. """
        return self.m_comp.get_parent()

    # Function: get_comp
    #
    # Returns a handle to the internal proxy component representing this port.
    #
    # Ports are considered components. However, they do not inherit
    # <uvm_component>. Instead, they contain an instance of
    # <uvm_port_component #(PORT)> that serves as a proxy to this port.
    def get_comp(self):
        return self.m_comp

    # Function: get_type_name
    #
    # Returns the type name to this port. Derived port classes must implement
    # this method to return the concrete type. Otherwise, only a generic
    # "uvm_port", "uvm_export" or "uvm_implementation" is returned.
    def get_type_name(self):
        if self.m_port_type == UVM_PORT:
            return "port"
        elif self.m_port_type == UVM_EXPORT:
            return "export"
        else:
            return "implementation"

    def max_size(self):
        """
        Function: max_size

        Returns the maximum number of implementation ports that must
        be connected to this port by the end_of_elaboration phase.
        Returns:
        """
        return self.m_max_size


    def min_size(self):
        """
        Function: min_size

        Returns the minimum number of implementation ports that must
        be connected to this port by the end_of_elaboration phase.
        Returns:
        """
        return self.m_min_size

    # Function: is_unbounded
    #
    # Returns 1 if this port has no maximum on the number of implementation
    # ports this port can connect to. A port is unbounded when the ~max_size~
    # argument in the constructor is specified as ~UVM_UNBOUNDED_CONNECTIONS~.
    def is_unbounded(self):
        return self.m_max_size == UVM_UNBOUNDED_CONNECTIONS

    # Function: is_port

    def is_port(self):
        return self.m_port_type == UVM_PORT

    # Function: is_export

    def is_export(self):
        return self.m_port_type == UVM_EXPORT

    # Function: is_imp
    #
    # Returns 1 if this port is of the type given by the method name,
    # 0 otherwise.
    def is_imp(self):
        return self.m_port_type == UVM_IMPLEMENTATION

    # Function: size
    #
    # Gets the number of implementation ports connected to this port. The value
    # is not valid before the end_of_elaboration phase, as port connections have
    # not yet been resolved.
    def size(self):
        return len(self.m_imp_list)

    def set_if(self, index=0):
        self.m_if = self.get_if(index)
        if self.m_if is not None:
            self.m_def_index = index

    def m_get_if_mask(self):
        return self.m_if_mask

    # Function: set_default_index
    #
    # Sets the default implementation port to use when calling an interface
    # method. This method should only be called on UVM_EXPORT types. The value
    # must not be set before the end_of_elaboration phase, when port connections
    # have not yet been resolved.
    def set_default_index(self, index):
        self.m_def_index = index

    def connect(self, provider):
        """
        Function: connect

        Connects this port to the given `provider` port. The ports must be
        compatible in the following ways

        - Their type parameters must match

        - The `provider`'s interface type (blocking, non-blocking, analysis, etc.)
            must be compatible. Each port has an interface mask that encodes the
            interface(s) it supports. If the bitwise AND of these masks is equal to
            the this port's mask, the requirement is met and the ports are
            compatible. For example, a uvm_blocking_put_port #(T) is compatible with
            a uvm_put_export #(T) and uvm_blocking_put_imp #(T) because the export
            and imp provide the interface required by the uvm_blocking_put_port.

        - Ports of type `UVM_EXPORT` can only connect to other exports or imps.

        - Ports of type `UVM_IMPLEMENTATION` cannot be connected, as they are
            bound to the component that implements the interface at time of
            construction.

        In addition to type-compatibility checks, the relationship between this
        port and the `provider` port will also be checked if the port's
        `check_connection_relationships` configuration has been set. (See `new`
        for more information.)

        Relationships, when enabled, are checked are as follows:

        - If this port is a UVM_PORT type, the `provider` can be a parent port,
            or a sibling export or implementation port.

        - If this port is a `UVM_EXPORT` type, the provider can be a child
            export or implementation port.

        If any relationship check is violated, a warning is issued.

        Note- the `UVMComponent.connect_phase` method is related to but not the same
        as this method. The component's `connect` method is a phase callback where
        port's `connect` method calls are made.
        Args:
            provider:
        """
        from .uvm_coreservice import UVMCoreService
        cs = UVMCoreService.get()
        top = cs.get_root()

        #if end_of_elaboration_ph.get_state() == UVM_PHASE_EXECUTING or end_of_elaboration_ph.get_state() == UVM_PHASE_DONE:
        #    self.m_comp.uvm_report_warning("Late Connection",
        #         ("Attempt to connect " + self.get_full_name()
        #             + " (of type " + self.get_type_name()
        #             + ") at or after end_of_elaboration phase. Ignoring."))
        #    return

        if provider is None:
            self.m_comp.uvm_report_error(s_connection_error_id,
                "Cannot connect to null port handle", UVM_NONE)
            return

        if provider == self:
            self.m_comp.uvm_report_error(s_connection_error_id,
               "Cannot connect a port instance to itself", UVM_NONE)
            return

        if (provider.m_if_mask & self.m_if_mask) != self.m_if_mask:
            print(sv.sformatf("provider: %h, m_if_mask: %h", provider.m_if_mask,
                    self.m_if_mask))
            self.m_comp.uvm_report_error(s_connection_error_id,
                 (provider.get_full_name()
                  + " (of type " + provider.get_type_name()
                  + ") doesn't provide the compl interface required of this port (type "
                  + self.get_type_name() + ")"), UVM_NONE)
            return

        # IMP.connect(anything) is illegal
        if self.is_imp() is True:
            parent_name = ""
            if self.m_comp.get_parent() is not None:
                parent_name = self.m_comp.get_full_name()
            self.m_comp.uvm_report_error(s_connection_error_id,
                 ERR_MSG2.format(parent_name, provider.get_full_name()), UVM_NONE)
            return

        # EXPORT.connect(PORT) are illegal
        if self.is_export() is True and provider.is_port() is True:
            self.m_comp.uvm_report_error(s_connection_error_id,
                 ERR_MSG3.format(provider.get_full_name()), UVM_NONE)
            return

        self.m_check_relationship(provider)

        self.m_provided_by[provider.get_full_name()] = provider
        provider.m_provided_to[self.get_full_name()] = self


    indent = ""
    save = ""

    # Function: debug_connected_to
    #
    # The ~debug_connected_to~ method outputs a visual text display of the
    # port/export/imp network to which this port connects (i.e., the port's
    # fanout).
    #
    # This method must not be called before the end_of_elaboration phase, as port
    # connections are not resolved until then.
    def debug_connected_to(self, level=0, max_level=-1):
        sz = 0
        num = 0
        curr_num = 0
        port = None
        # indent = ""

        if level < 0:
            level = 0
        if level == 0:
            UVMPortBase.save = ""
            UVMPortBase.indent = "    "

        if max_level != -1 and level >= max_level:
            return

        num = len(self.m_provided_by)

        if len(self.m_provided_by) != 0:
            for nm in self.m_provided_by:
                port = self.m_provided_by[nm]
                curr_num += 1
                UVMPortBase.save += UVMPortBase.indent + "    | \n"
                UVMPortBase.save += UVMPortBase.indent + "    |_" + nm + " (" + port.get_type_name() + ")\n"
                if num > 1 and curr_num != num:
                    UVMPortBase.indent += "    | "
                else:
                    UVMPortBase.indent += "        "
                port.debug_connected_to(level+1, max_level)
                #indent = indent.substr(0,indent.len()-4-1); TODO

        if level == 0:
            if UVMPortBase.save != "":
                UVMPortBase.save = ("This port's fanout network:\n\n    " + self.get_full_name()
                    + " (" + self.get_type_name() + ")\n" + UVMPortBase.save + "\n")
            if len(self.m_imp_list) == 0:
                from .uvm_coreservice import UVMCoreService
                cs = UVMCoreService.get()
                top = cs.get_root()
                if end_of_elaboration_ph is not None:
                    if (end_of_elaboration_ph.get_state() == UVM_PHASE_EXECUTING or
                            end_of_elaboration_ph.get_state() == UVM_PHASE_DONE):
                        UVMPortBase.save += "    Connected implementations: none\n"
                    else:
                        UVMPortBase.save += "    Connected implementations: not resolved until end-of-elab\n"
            else:
                UVMPortBase.save += "    Resolved implementation list:\n"
                for nm in self.m_imp_list:
                    port = self.m_imp_list[nm]
                    s_sz = str(sz)
                    UVMPortBase.save += (UVMPortBase.indent + s_sz + ": " + nm + " (" +
                        port.get_type_name() + ")\n")
                    sz += 1
            self.m_comp.uvm_report_info("debug_connected_to", UVMPortBase.save)
        return UVMPortBase.save

    # Function: debug_provided_to
    #
    # The ~debug_provided_to~ method outputs a visual display of the port/export
    # network that ultimately connect to this port (i.e., the port's fanin).
    #
    # This method must not be called before the end_of_elaboration phase, as port
    # connections are not resolved until then.
    def debug_provided_to(self, level=0, max_level=-1):
        nm = ""
        num = 0
        curr_num = 0
        port = None

        if level < 0:
            level = 0
        if level == 0:
            UVMPortBase.save = ""
            UVMPortBase.indent = "    "

        if max_level != -1 and level > max_level:
            return

        num = len(self.m_provided_to)

        if num != 0:
            for nm in self.m_provided_to:
                port = self.m_provided_to[nm]
                curr_num += 1
                UVMPortBase.save += UVMPortBase.indent + "    | \n"
                UVMPortBase.save += UVMPortBase.indent + "    |_" + nm + " (" + port.get_type_name() + ")\n"
                if num > 1 and curr_num != num:
                    UVMPortBase.indent += "    | "
                else:
                    UVMPortBase.indent += "        "
                port.debug_provided_to(level+1, max_level)
                # indent = indent.substr(0,indent.len()-4-1); #TODO

        if level == 0:
            if UVMPortBase.save != "":
                UVMPortBase.save = ("This port's fanin network:\n\n    "
                + self.get_full_name() + " (" + self.get_type_name()
                + ")\n" + UVMPortBase.save + "\n")
            if len(self.m_provided_to) == 0:
                UVMPortBase.save += UVMPortBase.indent + "This port has not been bound\n"
            self.m_comp.uvm_report_info("debug_provided_to", UVMPortBase.save)
        return UVMPortBase.save

    # get_connected_to
    # ----------------

    def get_connected_to(self, port_list):
        for name in self.m_provided_by:
            port = self.m_provided_by[name]
            port_list[name] = port.get_comp()

    # get_provided_to
    # ---------------

    def get_provided_to(self, port_list):
        for name in self.m_provided_to:
            port = self.m_provided_to[name]
            port_list[name] = port.get_comp()

    # m_check_relationship
    # --------------------

    def m_check_relationship(self, provider):
        s = ""
        res_from = None
        from_parent = None
        to_parent = None
        from_gparent = None
        to_gparent = None

        # Checks that the connection is between ports that are hierarchically
        # adjacent (up or down one level max, or are siblings),
        # and check for legal direction, requirer.connect(provider).

        # if we're an analysis port, allow connection to anywhere
        if self.get_type_name() == "uvm_analysis_port":
            return 1

        res_from        = self
        from_parent = self.get_parent()
        to_parent   = provider.get_parent()

        # skip check if we have a parentless port
        if from_parent is None or to_parent is None:
            return 1

        from_gparent = from_parent.get_parent()
        to_gparent     = to_parent.get_parent()

        # Connecting port-to-port: CHILD.port.connect(PARENT.port)
        #
        if res_from.is_port() and provider.is_port() and from_gparent != to_parent:
            s = (provider.get_full_name()
                + " (of type " + provider.get_type_name()
                + ") is not up one level of hierarchy from this port. "
                + "A port-to-port connection takes the form "
                + "child_component.child_port.connect(parent_port)")
            self.m_comp.uvm_report_warning(s_connection_warning_id, s, UVM_NONE)
            return 0

        # Connecting port-to-export: SIBLING.port.connect(SIBLING.export)
        # Connecting port-to-imp:        SIBLING.port.connect(SIBLING.imp)
        #
        elif res_from.is_port() and (provider.is_export() or provider.is_imp()) and from_gparent != to_gparent:
            s = (provider.get_full_name()
                + " (of type " + provider.get_type_name()
                + ") is not at the same level of hierarchy as this port. "
                + "A port-to-export connection takes the form "
                + "component1.port.connect(component2.export)")
            self.m_comp.uvm_report_warning(s_connection_warning_id, s, UVM_NONE)
            return 0

        # Connecting export-to-export: PARENT.export.connect(CHILD.export)
        # Connecting export-to-imp:        PARENT.export.connect(CHILD.imp)
        #
        elif res_from.is_export() and (provider.is_export() or provider.is_imp()) and from_parent != to_gparent:
            s = (provider.get_full_name()
                + " (of type " + provider.get_type_name()
                + ") is not down one level of hierarchy from this export. "
                + "An export-to-export or export-to-imp connection takes the form "
                + "parent_export.connect(child_component.child_export)")
            self.m_comp.uvm_report_warning(s_connection_warning_id, s, UVM_NONE)
            return 0
        return 1

    def m_add_list(self, provider):
        """
        m_add_list

        Internal method.
        Args:
            provider:
        """
        sz = ""
        for i in range(0, provider.size()):
            imp = provider.get_if(i)
            if not imp.get_full_name() in self.m_imp_list:
                self.m_imp_list[imp.get_full_name()] = imp

    # Function: resolve_bindings
    #
    # This callback is called just before entering the end_of_elaboration phase.
    # It recurses through each port's fanout to determine all the imp
    # destinations. It then checks against the required min and max connections.
    # After resolution, <size> returns a valid value and <get_if>
    # can be used to access a particular imp.
    #
    # This method is automatically called just before the start of the
    # end_of_elaboration phase. Users should not need to call it directly.
    def resolve_bindings(self):
        if self.m_resolved:  # don't repeat ourselves
            return

        if self.is_imp() is True:
            self.m_imp_list[self.get_full_name()] = self
        else:
            for nm in self.m_provided_by:
                port = self.m_provided_by[nm]
                port.resolve_bindings()
                self.m_add_list(port)

        self.m_resolved = 1

        if self.size() < self.min_size():
            self.m_comp.uvm_report_error(s_connection_error_id,
                "connection count of {} does not meet required minimum of {}"
                .format(self.size(), self.min_size()), UVM_NONE)

        if self.max_size() != UVM_UNBOUNDED_CONNECTIONS and self.size() > self.max_size():
            self.m_comp.uvm_report_error(s_connection_error_id,
                "connection count of {} exceeds maximum of {}".format(
                    self.size(), self.max_size()), UVM_NONE)

        if self.size() > 0:
            self.set_if(0)

    # Function: get_if
    #
    # Returns the implementation (imp) port at the given index from the array of
    # imps this port is connected to. Use <size> to get the valid range for index.
    # This method can only be called at the end_of_elaboration phase or after, as
    # port connections are not resolved before then.
    def get_if(self, index=0):
        s = ""
        if self.size() == 0:
            self.m_comp.uvm_report_warning("get_if",
                "Port size is zero; cannot get interface at any index", UVM_NONE)
            return None
        if index < 0 or index >= self.size():
            s = "Index {} out of range [0,{}]".format(index, self.size()-1)
            self.m_comp.uvm_report_warning(s_connection_error_id, s, UVM_NONE)
            return None
        for nm in self.m_imp_list:
            if index == 0:
                return self.m_imp_list[nm]
            index -= 1



#//
#// -------------------------------------------------------------
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

from uvm import *


class vip_tr(UVMSequenceItem):

    def __init__(self, name="vip_tr"):
        super().__init__(name)
        self.chr = 0x0
        self.rand('chr', range(0, 256))


    def convert2string(self):
        return sv.sformatf("0x%0h", self.chr)


uvm_object_utils_begin(vip_tr)
uvm_field_int('vip_tr', UVM_ALL_ON)
uvm_object_utils_end(vip_tr)



#//
#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc. 
#//   Copyright 2010 Synopsys, Inc.
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------
#
#//------------------------------------------------------------------------------
#//
#// CLASS: uvm_push_driver #(REQ,RSP)
#//
#// Base class for a driver that passively receives transactions, i.e. does not
#// initiate requests transactions. Also known as ~push~ mode. Its ports are
#// typically connected to the corresponding ports in a push sequencer as follows:
#//
#//|  push_sequencer.req_port.connect(push_driver.req_export);
#//|  push_driver.rsp_port.connect(push_sequencer.rsp_export);
#//
#// The ~rsp_port~ needs connecting only if the driver will use it to write
#// responses to the analysis export in the sequencer.
#//
#//------------------------------------------------------------------------------
#
#class uvm_push_driver #(type REQ=uvm_sequence_item,
#                        type RSP=REQ) extends uvm_component;
#
#  // Port: req_export
#  //
#  // This export provides the blocking put interface whose default 
#  // implementation produces an error. Derived drivers must override ~put~
#  // with an appropriate implementation (and not call super.put). Ports
#  // connected to this export will supply the driver with transactions.
#
#  uvm_blocking_put_imp #(REQ, uvm_push_driver #(REQ,RSP)) req_export;
#
#  // Port: rsp_port
#  //
#  // This analysis port is used to send response transactions back to the
#  // originating sequencer.
#
#  uvm_analysis_port #(RSP) rsp_port;
#
#  REQ req;
#  RSP rsp;
#
#  // Function: new
#  //
#  // Creates and initializes an instance of this class using the normal
#  // constructor arguments for <uvm_component>: ~name~ is the name of the
#  // instance, and ~parent~ is the handle to the hierarchical parent, if any.
#
#  function new (string name, uvm_component parent);
#    super.new(name, parent);
#    req_export = new("req_export", this);
#    rsp_port   = new("rsp_port", this);
#  endfunction
#
#  function void check_port_connections();
#    if (req_export.size() != 1)
#    uvm_report_fatal("Connection Error",
#                     $sformatf("Must connect to seq_item_port(%0d)",
#                               req_export.size()), UVM_NONE);
#  endfunction
#  
#  virtual function void end_of_elaboration_phase(uvm_phase phase);
#    super.end_of_elaboration_phase(phase);
#    check_port_connections();
#  endfunction
#  
#  virtual task put(REQ item);
#    uvm_report_fatal("UVM_PUSH_DRIVER", "Put task for push driver is not implemented", UVM_NONE);
#  endtask
#
#  const static string type_name = "uvm_push_driver #(REQ,RSP)";
#
#  virtual function string get_type_name ();
#    return type_name;
#  endfunction
#
#endclass
#



#
#//
#//----------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2013      NVIDIA Corporation
#//   Copyright 2019-2021 Tuomas Poikela
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
"""
Title: Objection Mechanism

The following classes define the objection mechanism and end-of-test
functionality, which is based on <uvm_objection>.
"""

import cocotb
from cocotb.triggers import Event, Timer
from .uvm_report_object import UVMReportObject
from .uvm_debug import uvm_debug
from .uvm_globals import *
from .uvm_object_globals import (UVM_RAISED, UVM_DROPPED, UVM_ALL_DROPPED)
from .sv import sv
from ..macros import (uvm_info, uvm_do_callbacks,
    uvm_do_callbacks_async)
from typing import List, Dict, Any
from .uvm_pool import UVMPool
from .uvm_callback import UVMCallback

UVM_USE_PROCESS_CONTAINER = 1


def classmethod_named(func):
    new_func = classmethod(func)
    setattr(new_func, '__name__', func.__name__)
    return new_func


class UVMObjectionEvents():
    def __init__(self):
        self.waiters = 0
        self.raised = Event('raised')
        self.dropped = Event('dropped')
        self.all_dropped = Event('all_dropped')


ObjContextDict = Dict[Any, 'UVMObjectionContextObject']
ObjContextList = List['UVMObjectionContextObject']


def get_name_depth(curr_obj_name: str) -> int:
    """ Returns the hier depth of the name """
    depth = 0
    for i in curr_obj_name:
        if i == ".":
            depth += 1
    return depth


def get_leaf_name(curr_obj_name: str) -> str:
    name = curr_obj_name
    rr = range(len(curr_obj_name)-1, -1, -1)
    for i in rr:
        if curr_obj_name[i] == ".":
            name = curr_obj_name[i+1:len(curr_obj_name)]
            break
    return name




class UVMObjectionCallback(UVMCallback):
    """
    The UVMObjection is the callback type that defines the callback
    implementations for an objection callback. A user uses the callback
    type uvm_objection_cbs_t to add callbacks to specific objections.

    .. code-block:: python

      class my_objection_cb extends uvm_objection_callback
        def __init__(self, name):
          super().__init__(name)

        def raised(objection, obj, source_obj, description, count):
            uvm_info("RAISED","{}: Objection {}: Raised for {}".format(sv.time(),
                objection.get_name(), obj.get_full_name())
      ...
      @cocotb.test()
      async def initial_begin(dut):
        cb = my_objection_cb ("cb")
        uvm_objection_cbs_t.add(None, cb) # typewide callback
    """
    def __init__(self, name):
        super().__init__(name)
    #  // Function: raised
    #  //
    #  // Objection raised callback function. Called by <uvm_objection::raised>.
    #
    #  virtual function void raised (uvm_objection objection, uvm_object obj,
    #      uvm_object source_obj, string description, int count)
    #  endfunction
    #
    #  // Function: dropped
    #  //
    #  // Objection dropped callback function. Called by <uvm_objection::dropped>.
    #
    #  virtual function void dropped (uvm_objection objection, uvm_object obj,
    #      uvm_object source_obj, string description, int count)
    #  endfunction
    #
    #  // Function: all_dropped
    #  //
    #  // Objection all_dropped callback function. Called by <uvm_objection::all_dropped>.
    #
    #  virtual task all_dropped (uvm_objection objection, uvm_object obj,
    #      uvm_object source_obj, string description, int count)
    #  endtask


class UVMObjection(UVMReportObject):
    """
    Class: UVMObjection

    Objections provide a facility for coordinating status information between
    two or more participating components, objects, and even module-based IP.

    Tracing of objection activity can be turned on to follow the activity of
    the objection mechanism. It may be turned on for a specific objection
    instance with <uvm_objection::trace_mode>, or it can be set for all
    objections from the command line using the option +UVM_OBJECTION_TRACE.
    """


    m_objections = []  # static uvm_objection[$]
    #  `uvm_register_cb(uvm_objection, uvm_objection_callback)
    #
    #  protected bit     self.m_trace_mode
    #  protected int     self.m_source_count[uvm_object]
    #  protected int     self.m_total_count [uvm_object]
    #  protected time    self.m_drain_time  [uvm_object]
    #  protected uvm_objection_events self.m_events [uvm_object]
    #  /*protected*/ bit     self.m_top_all_dropped
    #
    #  protected uvm_root self.m_top
    #
    #
    #  //// Drain Logic
    #
    #  // The context pool holds used context objects, so that
    #  // they're not constantly being recreated.  The maximum
    #  // number of contexts in the pool is equal to the maximum
    #  // number of simultaneous drains you could have occuring,
    #  // both pre and post forks.
    #  //
    #  // There's the potential for a programmability within the
    #  // library to dictate the largest this pool should be allowed
    #  // to grow, but that seems like overkill for the time being.
    #  local static uvm_objection_context_object UVMObjection.m_context_pool[$]
    m_context_pool = []

    # Notified when m_scheduled_list is not empty
    m_scheduled_list_not_empty_event = Event('m_scheduled_list_not_empty_event')

    #  // These are the active drain processes, which have been
    #  // forked off by the background process.  A raise can
    #  // use this array to kill a drain.
    #`ifndef UVM_USE_PROCESS_CONTAINER
    #  local process self.m_drain_proc[uvm_object]
    #`else
    #  local process_container_c self.m_drain_proc[uvm_object]
    #`endif

    #  // These are the contexts which have been scheduled for
    #  // retrieval by the background process, but which the
    #  // background process hasn't seen yet.
    #  local static uvm_objection_context_object m_scheduled_list[$]
    m_scheduled_list: ObjContextList = []

    #  // Once a context is seen by the background process, it is
    #  // removed from the scheduled list, and placed in the forked
    #  // list.  At the same time, it is placed in the scheduled
    #  // contexts array.  A re-raise can use the scheduled contexts
    #  // array to detect (and cancel) the drain.
    #  local uvm_objection_context_object self.m_scheduled_contexts[uvm_object]

    #  // Once the forked drain has actually started (this occurs
    #  // ~1 delta AFTER the background process schedules it), the
    #  // context is removed from the above array and list, and placed
    #  // in the forked_contexts list.
    #  local uvm_objection_context_object self.m_forked_contexts[uvm_object]
    #
    #  protected bit self.m_prop_mode = 1

    #  // Function: new
    #  //
    #  // Creates a new objection instance. Accesses the command line
    #  // argument +UVM_OBJECTION_TRACE to turn tracing on for
    #  // all objection objects.
    #
    def __init__(self, name=""):
        #uvm_cmdline_processor clp
        #uvm_coreservice_t cs_
        trace_args = []  # string [$]
        UVMReportObject.__init__(self, name)
        from .uvm_coreservice import UVMCoreService
        cs_ = UVMCoreService.get()
        #cs_ = uvm_coreservice_t::get()
        self.m_top = cs_.get_root()

        self.m_cleared = 0  # protected bit /* for checking obj count<0 */

        self.m_forked_list: ObjContextList = []  # uvm_objection_context_object[$]

        self.m_scheduled_contexts: ObjContextDict = {}  # uvm_objection_context_object[uvm_object]
        self.m_forked_contexts: ObjContextDict = {}  # uvm_objection_context_object[uvm_object]

        self.m_source_count = {}  # int[uvm_object]
        self.m_total_count = {}  # int[uvm_object]
        self.m_drain_time = {}  # time [uvm_object]
        self.m_events: Dict[Any, UVMObjectionEvents] = {}  # uvm_objection_events[uvm_object]
        self.m_top_all_dropped = 0

        self.m_drain_proc = {}  # process_container_c [uvm_object]

        self.set_report_verbosity_level(self.m_top.get_report_verbosity_level())

        # Get the command line trace mode setting
        #clp = uvm_cmdline_processor::get_inst()
        from .uvm_cmdline_processor import UVMCmdlineProcessor
        clp = UVMCmdlineProcessor.get_inst()

        self.m_prop_mode = 1
        self.m_trace_mode = 0
        if clp.get_arg_matches("+UVM_OBJECTION_TRACE", trace_args):
            self.m_trace_mode = 1
        UVMObjection.m_objections.append(self)

    def trace_mode(self, mode=-1):
        """
        Set or get the trace mode for the objection object. If no
        argument is specified (or an argument other than 0 or 1)
        the current trace mode is unaffected. A trace_mode of
        0 turns tracing off. A trace mode of 1 turns tracing on.
        The return value is the mode prior to being reset.

        Args:
            mode (int): 0=disable, 1=enable

        Returns int:
        """
        trace_mode = self.m_trace_mode
        if mode == 0:
            self.m_trace_mode = 0
        elif mode == 1:
            self.m_trace_mode = 1
        return trace_mode

    #  // Function- m_report
    #  //
    #  // Internal method for reporting count updates
    #
    def m_report(self, obj, source_obj, description: str, count: int, action: str) -> None:
        _count = 0
        if obj in self.m_source_count:
            _count = self.m_source_count[obj]
        _total = 0
        if obj in self.m_total_count:
            _total = self.m_total_count[obj]
        if (not uvm_report_enabled(UVM_NONE, UVM_INFO,"OBJTN_TRC") or (not self.m_trace_mode)):
            return

        descr = ""
        if description != "":
            descr = " (" + description + ")"

        if source_obj is obj:
            name = obj.get_full_name()
            if name == "":
                name = "uvm_top"
            uvm_report_info("OBJTN_TRC",
                sv.sformatf("Object %0s %0s %0d objection(s)%s: count=%0d  total=%0d",
                name, action, count, descr, _count, _total), UVM_NONE)
        else:
            cpath = 0
            last_dot = 0
            sname = source_obj.get_full_name()
            nm = obj.get_full_name()
            _max = sname.len()
            if sname.len() > nm.len():
                _max = nm.len()

            # For readability, only print the part of the source obj hierarchy underneath
            # the current object.
            while ((sname[cpath] == nm[cpath]) and (cpath < _max)):
                if (sname[cpath] == "."):
                    last_dot = cpath
                cpath += 1

            if last_dot:
                sname = sname.substr(last_dot+1, sname.len())

        name = obj.get_full_name()
        if name == "":
            name = "uvm_top"
        act_type = "subtracted"
        if action == "raised":
            act_type = "added"
        act_dir = "from"
        if action == "raised":
            act_dir = "to"
        uvm_report_info("OBJTN_TRC",
            sv.sformatf("Object %0s %0s %0d objection(s) %0s its total (%s from"
            " source object %s%s): count=%0d  total=%0d", name, act_type, count,
            act_dir, action, sname, descr, _count, _total), UVM_NONE)

    #  // Function- m_get_parent
    #  //
    #  // Internal method for getting the parent of the given ~object~.
    #  // The ultimate parent is uvm_top, UVM's implicit top-level component.
    #
    def m_get_parent(self, obj):
        comp = None
        seq = None
        if hasattr(obj, 'get_parent'):
            comp = obj
            obj = comp.get_parent()
        elif hasattr(obj, 'get_sequencer'):
            seq = obj
            obj = seq.get_sequencer()
        else:
            obj = self.m_top
        if obj is None:
            obj = self.m_top
        return obj


    #  // Function- m_propagate
    #  //
    #  // Propagate the objection to the objects parent. If the object is a
    #  // component, the parent is just the hierarchical parent. If the object is
    #  // a sequence, the parent is the parent sequence if one exists, or
    #  // it is the attached sequencer if there is no parent sequence.
    #  //
    #  // obj : the uvm_object on which the objection is being raised or lowered
    #  // source_obj : the root object on which the end user raised/lowered the
    #  //   objection (as opposed to an anscestor of the end user object)a
    #  // count : the number of objections associated with the action.
    #  // raise : indicator of whether the objection is being raised or lowered. A
    #  //   1 indicates the objection is being raised.
    #
    def m_propagate(self, obj, source_obj, description, count, raise_, in_top_thread):
        if obj is not None and obj != self.m_top:
            parent_obj = self.m_get_parent(obj)
            if raise_:
                self.m_raise(parent_obj, source_obj, description, count)
            else:
                self.m_drop(parent_obj, source_obj, description, count, in_top_thread)

    #  // Group: Objection Control
    #
    #  // Function: set_propagate_mode
    #  // Sets the propagation mode for this objection.
    #  //
    #  // By default, objections support hierarchical propagation for
    #  // components.  For example, if we have the following basic
    #  // component tree:
    #  //
    #  //| uvm_top.parent.child
    #  //
    #  // Any objections raised by 'child' would get propagated
    #  // down to parent, and then to uvm_test_top.  Resulting in the
    #  // following counts and totals:
    #  //
    #  //|                      | count | total |
    #  //| uvm_top.parent.child |     1 |    1  |
    #  //| uvm_top.parent       |     0 |    1  |
    #  //| uvm_top              |     0 |    1  |
    #  //|
    #  //
    #  // While propagations such as these can be useful, if they are
    #  // unused by the testbench then they are simply an unnecessary
    #  // performance hit.  If the testbench is not going to use this
    #  // functionality, then the performance can be improved by setting
    #  // the propagation mode to 0.
    #  //
    #  // When propagation mode is set to 0, all intermediate callbacks
    #  // between the ~source~ and ~top~ will be skipped.  This would
    #  // result in the following counts and totals for the above objection:
    #  //
    #  //|                      | count | total |
    #  //| uvm_top.parent.child |     1 |    1  |
    #  //| uvm_top.parent       |     0 |    0  |
    #  //| uvm_top              |     0 |    1  |
    #  //|
    #  //
    #  // Since the propagation mode changes the behavior of the objection,
    #  // it can only be safely changed if there are no objections ~raised~
    #  // or ~draining~.  Any attempts to change the mode while objections
    #  // are ~raised~ or ~draining~ will result in an error.
    #  //
    #  function void set_propagate_mode (bit prop_mode)
    #     if (!self.m_top_all_dropped && (get_objection_total() != 0)) begin
    #        `uvm_error("UVM/BASE/OBJTN/PROP_MODE",
    #                   {"The propagation mode of '", this.get_full_name(),
    #                    "' cannot be changed while the objection is raised ",
    #                    "or draining!"})
    #        return
    #     end
    #
    #     self.m_prop_mode = prop_mode
    #  endfunction : set_propagate_mode

    #  // Function: get_propagate_mode
    #  // Returns the propagation mode for this objection.
    def get_propagate_mode(self) -> int:
        return self.m_prop_mode

    #  // Function: raise_objection
    #  //
    #  // Raises the number of objections for the source ~object~ by ~count~, which
    #  // defaults to 1.  The ~object~ is usually the ~this~ handle of the caller.
    #  // If ~object~ is not specified or ~null~, the implicit top-level component,
    #  // <uvm_root>, is chosen.
    #  //
    #  // Raising an objection causes the following.
    #  //
    #  // - The source and total objection counts for ~object~ are increased by
    #  //   ~count~. ~description~ is a string that marks a specific objection
    #  //   and is used in tracing/debug.
    #  //
    #  // - The objection's <raised> virtual method is called, which calls the
    #  //   <uvm_component::raised> method for all of the components up the
    #  //   hierarchy.
    #  //
    def raise_objection(self, obj=None, description="", count=1):
        if obj is None:
            obj = self.m_top
        self.m_cleared = 0
        self.m_top_all_dropped = 0
        uvm_debug(self, 'raise_objection', obj.get_name() + " Starting to raise objection")
        self.m_raise(obj, obj, description, count)

    #  // Function- m_raise
    def m_raise(self, obj, source_obj, description="", count=1):
        idx = 0
        ctxt = None  # uvm_objection_context_object

        # Ignore raise if count is 0
        if count == 0:
            return

        if obj in self.m_total_count:
            self.m_total_count[obj] += count
        else:
            self.m_total_count[obj] = count

        if source_obj == obj:
            if obj in self.m_source_count:
                self.m_source_count[obj] += count
            else:
                self.m_source_count[obj] = count

        if self.m_trace_mode:
            self.m_report(obj,source_obj,description,count,"raised")

        self.raised(obj, source_obj, description, count)

        # Handle any outstanding drains...
        # First go through the scheduled list
        idx = 0
        m_scheduled_list = UVMObjection.m_scheduled_list
        # while idx < len(m_scheduled_list):
        for entry in m_scheduled_list:
            if ((entry.obj == obj) and
                    (entry.objection == self)):
                # Caught it before the drain was forked
                ctxt = entry
                del m_scheduled_list[idx]
                break
            idx += 1

        # If it's not there, go through the forked list
        if ctxt is None:
            idx = 0
            while idx < len(self.m_forked_list):
                if self.m_forked_list[idx].obj == obj:
                    # Caught it after the drain was forked,
                    # but before the fork started
                    ctxt = self.m_forked_list[idx]
                    del self.m_forked_list[idx]
                    del self.m_scheduled_contexts[ctxt.obj]
                    break
                idx += 1

        # If it's not there, go through the forked contexts
        if ctxt is None:
            if obj in self.m_forked_contexts:
                # Caught it with the forked drain running
                ctxt = self.m_forked_contexts[obj]
                del self.m_forked_contexts[obj]
                # Kill the drain

        uvm_debug(self, 'm_raise', obj.get_name() + " ENDING FUNC")
        # TODO
        #if UVM_USE_PROCESS_CONTAINER:
        #    self.m_drain_proc[obj].kill()
        #    del self.m_drain_proc[obj]
        #else:
        #    self.m_drain_proc[obj].p.kill()
        #    del self.m_drain_proc[obj]
        uvm_debug(self, 'm_raise', obj.get_name() + " NEVER GETS HERE")

        if ctxt is None:
            # If there were no drains, just propagate as usual
            if not self.m_prop_mode and obj != self.m_top:
                uvm_debug(self, 'm_raise', obj.get_name() + " XXX NEVER GETS HERE")
                self.m_raise(self.m_top,source_obj,description,count)
            elif obj != self.m_top:
                self.m_propagate(obj, source_obj, description, count, 1, 0)

        else:
            # Otherwise we need to determine what exactly happened

            # Determine the diff count, if it's positive, then we're
            # looking at a 'raise' total, if it's negative, then
            # we're looking at a 'drop', but not down to 0.  If it's
            # a 0, that means that there is no change in the total.
            diff_count = count - ctxt.count

            if diff_count != 0:
                # Something changed
                if diff_count > 0:
                    # we're looking at an increase in the total
                    if not self.m_prop_mode and obj != self.m_top:
                        self.m_raise(self.m_top, source_obj, description, diff_count)
                    elif obj != self.m_top:
                        self.m_propagate(obj, source_obj, description, diff_count, 1, 0)
                else:
                    # we're looking at a decrease in the total
                    # The count field is always positive...
                    diff_count = -diff_count
                    if not self.m_prop_mode and obj != self.m_top:
                        self.m_drop(self.m_top, source_obj, description, diff_count)
                    elif obj != self.m_top:
                        self.m_propagate(obj, source_obj, description, diff_count, 0, 0)

            # Cleanup
            ctxt.clear()
            UVMObjection.m_context_pool.append(ctxt)
        #  endfunction

    #  // Function: drop_objection
    #  //
    #  // Drops the number of objections for the source ~object~ by ~count~, which
    #  // defaults to 1.  The ~object~ is usually the ~this~ handle of the caller.
    #  // If ~object~ is not specified or ~null~, the implicit top-level component,
    #  // <uvm_root>, is chosen.
    #  //
    #  // Dropping an objection causes the following.
    #  //
    #  // - The source and total objection counts for ~object~ are decreased by
    #  //   ~count~. It is an error to drop the objection count for ~object~ below
    #  //   zero.
    #  //
    #  // - The objection's <dropped> virtual method is called, which calls the
    #  //   <uvm_component::dropped> method for all of the components up the
    #  //   hierarchy.
    #  //
    #  // - If the total objection count has not reached zero for ~object~, then
    #  //   the drop is propagated up the object hierarchy as with
    #  //   <raise_objection>. Then, each object in the hierarchy will have updated
    #  //   their ~source~ counts--objections that they originated--and ~total~
    #  //   counts--the total number of objections by them and all their
    #  //   descendants.
    #  //
    #  // If the total objection count reaches zero, propagation up the hierarchy
    #  // is deferred until a configurable drain-time has passed and the
    #  // <uvm_component::all_dropped> callback for the current hierarchy level
    #  // has returned. The following process occurs for each instance up
    #  // the hierarchy from the source caller:
    #  //
    #  // A process is forked in a non-blocking fashion, allowing the ~drop~
    #  // call to return. The forked process then does the following:
    #  //
    #  // - If a drain time was set for the given ~object~, the process waits for
    #  //   that amount of time.
    #  //
    #  // - The objection's <all_dropped> virtual method is called, which calls the
    #  //   <uvm_component::all_dropped> method (if ~object~ is a component).
    #  //
    #  // - The process then waits for the ~all_dropped~ callback to complete.
    #  //
    #  // - After the drain time has elapsed and all_dropped callback has
    #  //   completed, propagation of the dropped objection to the parent proceeds
    #  //   as described in <raise_objection>, except as described below.
    #  //
    #  // If a new objection for this ~object~ or any of its descendants is raised
    #  // during the drain time or during execution of the all_dropped callback at
    #  // any point, the hierarchical chain described above is terminated and the
    #  // dropped callback does not go up the hierarchy. The raised objection will
    #  // propagate up the hierarchy, but the number of raised propagated up is
    #  // reduced by the number of drops that were pending waiting for the
    #  // all_dropped/drain time completion. Thus, if exactly one objection
    #  // caused the count to go to zero, and during the drain exactly one new
    #  // objection comes in, no raises or drops are propagated up the hierarchy,
    #  //
    #  // As an optimization, if the ~object~ has no set drain-time and no
    #  // registered callbacks, the forked process can be skipped and propagation
    #  // proceeds immediately to the parent as described.
    #
    def drop_objection(self, obj=None, description="", count=1):
        if obj is None:
            obj = self.m_top
        uvm_debug(self, 'drop_objection', obj.get_name() + " Starting to drop objection")

        self.m_drop(obj, obj, description, count, 0)

    #  // Function- m_drop
    #
    def m_drop(self, obj, source_obj, description="", count=1,
            in_top_thread=0):
        # Ignore drops if the count is 0
        if count == 0:
            return

        if obj not in self.m_total_count or (count > self.m_total_count[obj]):
            if self.m_cleared:
                return
            uvm_report_fatal("OBJTN_ZERO", ("Object '" + obj.get_full_name()
              + "' attempted to drop total objection '" + self.get_name() + "' count below zero. " +
              "Description: '" + description + "' source_obj: " +
              source_obj.get_name()))
            return

        if obj == source_obj:
            if obj not in self.m_source_count or (count > self.m_source_count[obj]):
                if self.m_cleared:
                    return
                uvm_report_fatal("OBJTN_ZERO", ("Object \"" + obj.get_full_name()
                  + "\" attempted to drop source objection '" + self.get_name() + "' count below zero"))
                return
            self.m_source_count[obj] -= count

        self.m_total_count[obj] -= count

        if self.m_trace_mode:
            self.m_report(obj,source_obj,description,count,"dropped")

        self.dropped(obj, source_obj, description, count)

        # if count != 0, no reason to fork
        if self.m_total_count[obj] != 0:
            if not self.m_prop_mode and obj != self.m_top:
                self.m_drop(self.m_top,source_obj,description, count, in_top_thread)
            elif obj != self.m_top:
                self.m_propagate(obj, source_obj, description, count, 0, in_top_thread)
        else:
            ctxt = None  # uvm_objection_context_object
            if (len(UVMObjection.m_context_pool) > 0):
                ctxt = UVMObjection.m_context_pool.pop(0)
            else:
                ctxt = UVMObjectionContextObject()

            ctxt.obj = obj
            ctxt.source_obj = source_obj
            ctxt.description = description
            ctxt.count = count
            ctxt.objection = self
            # Need to be thread-safe, let the background
            # process handle it.

            # Why don't we look at in_top_thread here?  Because
            # a re-raise will kill the drain at object that it's
            # currently occuring at, and we need the leaf-level kills
            # to not cause accidental kills at branch-levels in
            # the propagation.

            # Using the background process just allows us to
            # separate the links of the chain.
            UVMObjection.m_scheduled_list.append(ctxt)
            UVMObjection.m_scheduled_list_not_empty_event.set()
        #end // else: !if(self.m_total_count[obj] != 0)


    # Function: clear
    #
    # Immediately clears the objection state. All counts are cleared and the
    # any processes waiting on a call to wait_for(UVM_ALL_DROPPED, uvm_top)
    # are released.
    #
    # The caller, if a uvm_object-based object, should pass its 'this' handle
    # to the ~obj~ argument to document who cleared the objection.
    # Any drain_times set by the user are not affected.
    #
    def clear(self, obj=None):
        name = ""
        ctxt = None  # uvm_objection_context_object
        idx = 0

        uvm_debug(self, 'clear', 'START')
        if obj is None:
            obj = self.m_top

        name = obj.get_full_name()
        if name == "":
            name = "uvm_top"

        if (self.m_top_all_dropped is False and
                self.get_objection_total(self.m_top) > 0):
            uvm_report_warning("OBJTN_CLEAR",("Object '" + name
                + "' cleared objection counts for " + self.get_name()))
        # Should there be a warning if there are outstanding objections
        self.m_source_count = {}
        self.m_total_count = {}

        # Remove any scheduled drains from the static queue
        idx = 0
        while (idx < len(UVMObjection.m_scheduled_list)):
            if (UVMObjection.m_scheduled_list[idx].objection == self):
                UVMObjection.m_scheduled_list[idx].clear()
                UVMObjection.m_context_pool.append(UVMObjection.m_scheduled_list[idx])
                del UVMObjection.m_scheduled_list[idx]
            else:
                idx += 1

        # Scheduled contexts and m_forked_lists have duplicate
        # entries... clear out one, free the other.
        self.m_scheduled_contexts = {}
        while len(self.m_forked_list) > 0:
            self.m_forked_list[0].clear()
            UVMObjection.m_context_pool.append(self.m_forked_list[0])
            front = self.m_forked_list[0]
            self.m_forked_list[0].remove(front)

        # running drains have a context and a process
        for o in self.m_forked_contexts:
            if UVM_USE_PROCESS_CONTAINER:
                self.m_drain_proc[o].kill()
                del self.m_drain_proc[o]
            else:
                self.m_drain_proc[o].p.kill()
                del self.m_drain_proc[o]

            self.m_forked_contexts[o].clear()
            UVMObjection.m_context_pool.append(self.m_forked_contexts[o])
            del self.m_forked_contexts[o]

        self.m_top_all_dropped = False
        self.m_cleared = 1
        if self.m_top in self.m_events:
            self.m_events[self.m_top].all_dropped.set()

    #  // m_execute_scheduled_forks
    #  // -------------------------
    #  // background process; when non

    @classmethod
    async def m_execute_scheduled_forks(cls):
        while True:
            #wait(UVMObjection.m_scheduled_list.size() != 0)
            uvm_debug(cls, 'm_execute_scheduled_forks', 'waiting list to not be empty')
            await UVMObjection.m_scheduled_list_not_empty_event.wait()
            UVMObjection.m_scheduled_list_not_empty_event.clear()

            if len(UVMObjection.m_scheduled_list) != 0:
                # Save off the context before the fork
                c = UVMObjection.m_scheduled_list.pop(0)

                # A re-raise can use this to figure out props (if any)
                objection = c.objection
                #rm if objection is not None:
                objection.m_scheduled_contexts[c.obj] = c
                # The fork below pulls out from the forked list
                objection.m_forked_list.append(c)
                # The fork will guard the m_forked_drain call, but
                # a re-raise can kill self.m_forked_list contexts in the delta
                # before the fork executes.
                pproc = cocotb.fork(cls.m_execute_scheduled_forks_fork_join_none(c))
                #else:
                #    uvm_error("UVMObjection", "Null objection in objection context")


    @classmethod
    async def m_execute_scheduled_forks_fork_join_none(cls,
            c: 'UVMObjectionContextObject'):
        objection = c.objection  # automatic uvm_objection
        # Check to maike sure re-raise didn't empty the fifo
        uvm_debug(cls, 'm_execute_scheduled_forks_fork_join_none', 'check list len')
        if len(objection.m_forked_list) > 0:
            #uvm_objection_context_object ctxt
            ctxt = objection.m_forked_list.pop(0)
            # Clear it out of scheduled
            del objection.m_scheduled_contexts[ctxt.obj]

            # Move it in to forked (so re-raise can figure out props)
            objection.m_forked_contexts[ctxt.obj] = ctxt
            # Save off our process handle, so a re-raise can kill it...

            # TODO
            #if UVM_USE_PROCESS_CONTAINER == 0:
            #    objection.m_drain_proc[ctxt.obj] = process::self()
            #else:
            #    process_container_c c = new(process::self())
            #    objection.m_drain_proc[ctxt.obj]=c

            # Execute the forked drain
            await objection.m_forked_drain(ctxt.obj, ctxt.source_obj, ctxt.description,
                    ctxt.count, 1)
            # Cleanup if we survived (no re-raises)
            if ctxt.obj in objection.m_drain_proc:
                del objection.m_drain_proc[ctxt.obj]
            else:
                pass

            # tpoikela: Added check since ctxt.obj becomes None
            #if ctxt.obj in objection.m_forked_contexts:
            del objection.m_forked_contexts[ctxt.obj]
            # Clear out the context object (prevent memory leaks)
            ctxt.clear()
            # Save the context in the pool for later reuse
            UVMObjection.m_context_pool.append(ctxt)
        #rm await uvm_zero_delay()

    #  // m_forked_drain
    #  // -------------
    #  task m_forked_drain (uvm_object obj,
    #                       uvm_object source_obj,
    #                       string description="",
    #                       int count=1,
    #                       int in_top_thread=0)

    async def m_forked_drain(self, obj, source_obj, description="", count=1,
            in_top_thread=0):
        #rm diff_count = 0

        if obj in self.m_drain_time:
            # pass
            # TODO `uvm_delay(self.m_drain_time[obj])
            await Timer(self.m_drain_time[obj])

        if self.m_trace_mode:
            self.m_report(obj,source_obj,description,count,"all_dropped")

        await self.all_dropped(obj,source_obj,description, count)

        # wait for all_dropped cbs to complete
        await uvm_zero_delay()
        # TODO wait fork

        # we are ready to delete the 0-count entries for the current
        # object before propagating up the hierarchy.
        if obj in self.m_source_count and self.m_source_count[obj] == 0:
            del self.m_source_count[obj]

        if obj in self.m_total_count and self.m_total_count[obj] == 0:
            del self.m_total_count[obj]

        if not self.m_prop_mode and obj != self.m_top:
            self.m_drop(self.m_top,source_obj,description, count, 1)
        elif obj != self.m_top:
            self.m_propagate(obj, source_obj, description, count, 0, 1)

    #  // m_init_objections
    #  // -----------------
    #
    #  // Forks off the single background process
    @classmethod
    async def m_init_objections(cls):
        #uvm_debug(cls, 'm_init_objections', "Forking m_execute_scheduled_forks")
        pproc = cocotb.fork(cls.m_execute_scheduled_forks())
        #await uvm_zero_delay()

    #  // Function: set_drain_time
    #  //
    #  // Sets the drain time on the given ~object~ to ~drain~.
    #  //
    #  // The drain time is the amount of time to wait once all objections have
    #  // been dropped before calling the all_dropped callback and propagating
    #  // the objection to the parent.
    #  //
    #  // If a new objection for this ~object~ or any of its descendants is raised
    #  // during the drain time or during execution of the all_dropped callbacks,
    #  // the drain_time/all_dropped execution is terminated.
    #
    #  // AE: set_drain_time(drain,obj=null)?
    def set_drain_time(self, obj=None, drain=0):
        if obj is None:
            obj = self.m_top
            self.m_drain_time[obj] = drain

    #  //----------------------
    #  // Group: Callback Hooks
    #  //----------------------

    #  // Function: raised
    #  //
    #  // Objection callback that is called when a <raise_objection> has reached ~obj~.
    #  // The default implementation calls <uvm_component::raised>.
    #
    def raised(self, obj, source_obj, description, count):
        if hasattr(obj, 'raised'):
            obj.raised(self, source_obj, description, count)
        # TODO `uvm_do_callbacks(uvm_objection,uvm_objection_callback,raised(this,obj,source_obj,description,count))
        uvm_do_callbacks(self, UVMObjectionCallback, 'raised', self,obj,source_obj,description,count)
        if obj in self.m_events:
            self.m_events[obj].raised.set()

    #  // Function: dropped
    #  //
    #  // Objection callback that is called when a <drop_objection> has reached ~obj~.
    #  // The default implementation calls <uvm_component::dropped>.
    #
    #  virtual function void dropped (uvm_object obj,
    #                                 uvm_object source_obj,
    #                                 string description,
    #                                 int count)
    def dropped(self, obj, source_obj, description, count):
        comp = obj
        comp.dropped(self, source_obj, description, count)
        # TODO `uvm_do_callbacks(uvm_objection,uvm_objection_callback,dropped(this,obj,source_obj,description,count))
        uvm_do_callbacks(self,UVMObjectionCallback,'dropped', self,obj,source_obj,description,count)
        if obj in self.m_events:
            self.m_events[obj].dropped.set()


    async def all_dropped(self, obj, source_obj, description, count):
        """
        Objection callback that is called when a `drop_objection` has reached ~obj~,
        and the total count for ~obj~ goes to zero. This callback is executed
        after the drain time associated with ~obj~. The default implementation
        calls `UVMComponent.all_dropped`.

        Args:
            obj (UVMObject):
            source_obj (UVMObject):
            description (str):
            count (int):
        """
        comp = obj
        await comp.all_dropped(self, source_obj, description, count)
        # TODO `uvm_do_callbacks(uvm_objection,uvm_objection_callback,all_dropped(this,obj,source_obj,description,count))
        await uvm_do_callbacks_async(self, UVMObjectionCallback, 'all_dropped', self,obj, source_obj, description, count)
        if obj in self.m_events:
            self.m_events[obj].all_dropped.set()
        if obj == self.m_top:
            self.m_top_all_dropped = 1
    #  endtask

    #  //------------------------
    #  // Group: Objection Status
    #  //------------------------
    #
    #  // Function: get_objectors
    #  //
    #  // Returns the current list of objecting objects (objects that
    #  // raised an objection but have not dropped it).
    #
    #  function void get_objectors(ref uvm_object list[$])
    #    list.delete()
    #    foreach (self.m_source_count[obj]) list.push_back(obj)
    #  endfunction

    #  // Task: wait_for
    #  //
    #  // Waits for the raised, dropped, or all_dropped ~event~ to occur in
    #  // the given ~obj~. The task returns after all corresponding callbacks
    #  // for that event have been executed.
    #  //
    #  task wait_for(uvm_objection_event objt_event, uvm_object obj=null)
    async def wait_for(self, objt_event, obj=None):
        if obj is None:
            obj = self.m_top

        if (obj in self.m_events) is False:
            self.m_events[obj] = UVMObjectionEvents()

        self.m_events[obj].waiters += 1
        if objt_event == UVM_RAISED:
            await self.m_events[obj].raised.wait()
        elif objt_event == UVM_DROPPED:
            await self.m_events[obj].dropped.wait()
        elif objt_event == UVM_ALL_DROPPED:
            await self.m_events[obj].all_dropped.wait()

        self.m_events[obj].waiters -= 1

        if self.m_events[obj].waiters == 0:
            del self.m_events[obj]

    #
    #   task wait_for_total_count(uvm_object obj=null, int count=0)
    #     if (obj==null)
    #       obj = self.m_top
    #
    #     if(!self.m_total_count.exists(obj) && count == 0)
    #       return
    #     if (count == 0)
    #        wait (!self.m_total_count.exists(obj) && count == 0)
    #     else
    #        wait (self.m_total_count.exists(obj) && self.m_total_count[obj] == count)
    #   endtask


    def get_objection_count(self, obj=None):
        """
        Returns the current number of objections raised by the given ~object~.
        """
        if obj is None:
            obj = self.m_top

        if obj not in self.m_source_count:
            return 0
        return self.m_source_count[obj]

    def get_objection_total(self, obj=None):
        """
        Returns the current number of objections raised by the given ~object~
        and all descendants.
        """
        uvm_debug(self, 'get_objection_total', 'START')
        if obj is None:
            obj = self.m_top

        if obj not in self.m_total_count:
            uvm_debug(self, 'get_objection_total', 'Returning 0')
            return 0
        else:
            cnt = self.m_total_count[obj]
            uvm_debug(self, 'get_objection_total', 'Returning cnt ' + str(cnt))
            return self.m_total_count[obj]


    #  // Function: get_drain_time
    #  //
    #  // Returns the current drain time set for the given ~object~ (default: 0 ns).
    #
    #  function time get_drain_time (uvm_object obj=null)
    #    if (obj==null)
    #      obj = self.m_top
    #
    #    if (!self.m_drain_time.exists(obj))
    #      return 0
    #    return self.m_drain_time[obj]
    #  endfunction
    #
    #

    #  // m_display_objections
    #
    def m_display_objections(self, obj=None, show_header=1):
        blank = "                                                                                   "
        s = ""
        total = 0
        lst = UVMPool()
        curr_obj = None
        depth = 0
        name = ""
        this_obj_name = ""
        curr_obj_name = ""

        for o in self.m_total_count:
            theobj = o
            if self.m_total_count[o] > 0:
                lst[theobj.get_full_name()] = theobj

        if obj is None:
            obj = self.m_top

        total = self.get_objection_total(obj)
        s = sv.sformatf("The total objection count is %0d\n",total)

        if total == 0:
            return s

        s = s + "---------------------------------------------------------\n"
        s = s + "Source  Total   \n"
        s = s + "Count   Count   Object\n"
        s = s + "---------------------------------------------------------\n"

        this_obj_name = obj.get_full_name()
        curr_obj_name = this_obj_name

        while True:
            curr_obj = lst[curr_obj_name]

            # determine depth, count dots "." in the name
            depth = get_name_depth(curr_obj_name)

            # determine leaf name
            name = get_leaf_name(curr_obj_name)

            if curr_obj_name == "":
                name = "uvm_top"
            else:
                depth += 1

            # print it
            src_count = 0
            if curr_obj in self.m_source_count:
                src_count = self.m_source_count[curr_obj]
            total_count = 0
            if curr_obj in self.m_total_count:
                total_count = self.m_total_count[curr_obj]
            s += sv.sformatf("%d       %d       %s'%s'\n", src_count, total_count,
                    blank[0:2*depth], name)

            if not ((lst.has_next() and
                    curr_obj_name[0:len(this_obj_name)] == this_obj_name)):
                break
            curr_obj_name = lst.next()

        s = s + "---------------------------------------------------------\n"

        return s



    def convert2string(self):
        return self.m_display_objections(self.m_top,1)

    #
    #  // Function: display_objections
    #  //
    #  // Displays objection information about the given ~object~. If ~object~ is
    #  // not specified or ~null~, the implicit top-level component, <uvm_root>, is
    #  // chosen. The ~show_header~ argument allows control of whether a header is
    #  // output.
    def display_objections(self, obj=None, show_header=1):
        m = self.m_display_objections(obj,show_header)
        uvm_info("UVM/OBJ/DISPLAY", m, UVM_NONE)

    #
    #  // Below is all of the basic data stuff that is needed for a uvm_object
    #  // for factory registration, printing, comparing, etc.
    #
    #  typedef uvm_object_registry#(uvm_objection,"uvm_objection") type_id
    #  static function type_id get_type()
    #    return type_id::get()
    #  endfunction
    #
    #  function uvm_object create (string name="")
    #    uvm_objection tmp = new(name)
    #    return tmp
    #  endfunction
    #
    #  virtual function string get_type_name ()
    #    return "uvm_objection"
    #  endfunction

    def do_copy(self, rhs):
        _rhs = rhs
        # $cast(_rhs, rhs)
        self.m_source_count = _rhs.m_source_count
        self.m_total_count  = _rhs.m_total_count
        self.m_drain_time   = _rhs.m_drain_time
        self.m_prop_mode    = _rhs.self.m_prop_mode


#// TODO: change to plusarg
#//`define UVM_DEFAULT_TIMEOUT 9200s

#//------------------------------------------------------------------------------
#//
#// Class- uvm_test_done_objection DEPRECATED
#//
#// Provides built-in end-of-test coordination
#//------------------------------------------------------------------------------

#class uvm_test_done_objection extends uvm_objection
    #
    #   protected static uvm_test_done_objection m_inst
    #  protected bit m_forced
    #
    #  // For communicating all objections dropped and end of phasing
    #  local  bit m_executing_stop_processes
    #  local  int m_n_stop_threads
    #
    #
    #  // Function- new DEPRECATED
    #  //
    #  // Creates the singleton test_done objection. Users must not call
    #  // this method directly.
    #
    #  function new(string name="uvm_test_done")
    #    super.new(name)
    #  endfunction
    #
    #
    #  // Function- qualify DEPRECATED
    #  //
    #  // Checks that the given ~object~ is derived from either <uvm_component> or
    #  // <uvm_sequence_base>.
    #
    #  virtual function void qualify(uvm_object obj=null,
    #                                bit is_raise,
    #                                string description)
    #    uvm_component c
    #    uvm_sequence_base s
    #    string nm = is_raise ? "raise_objection" : "drop_objection"
    #    string desc = description == "" ? "" : {" (\"", description, "\")"}
    #    if(! ($cast(c,obj) || $cast(s,obj))) begin
    #      uvm_report_error("TEST_DONE_NOHIER", {"A non-hierarchical object, '",
    #        obj.get_full_name(), "' (", obj.get_type_name(),") was used in a call ",
    #        "to uvm_test_done.", nm,"(). For this objection, a sequence ",
    #        "or component is required.", desc })
    #    end
    #  endfunction
    #
    #  // Below are basic data operations needed for all uvm_objects
    #  // for factory registration, printing, comparing, etc.
    #
    #  typedef uvm_object_registry#(uvm_test_done_objection,"uvm_test_done") type_id
    #  static function type_id get_type()
    #    return type_id::get()
    #  endfunction
    #
    #  function uvm_object create (string name="")
    #    uvm_test_done_objection tmp = new(name)
    #    return tmp
    #  endfunction
    #
    #  virtual function string get_type_name ()
    #    return "uvm_test_done"
    #  endfunction
    #
    #  static function uvm_test_done_objection get()
    #    if(m_inst == null)
    #      m_inst = uvm_test_done_objection::type_id::create("run")
    #    return m_inst
    #  endfunction
    #
    #endclass

#// Have a pool of context objects to use
#class uvm_objection_context_object
class UVMObjectionContextObject:

    def __init__(self):
        """
           uvm_object obj
           uvm_object source_obj
           string description
           int    count
           uvm_objection objection
        """
        self.obj = None
        self.source_obj = None
        self.description = ""
        self.count = 0
        #self.objection: Optional[UVMObjection] = None
        self.objection = UVMObjection()

    def clear(self):
        """
        Clears the values stored within the object,
        preventing memory leaks from reused objects
        """
        self.obj = None
        self.source_obj = None
        self.description = ""
        self.count = 0
        self.objection = UVMObjection()


#// Typedef - Exists for backwards compat
#typedef uvm_objection uvm_callbacks_objection




#`endif
#



#//
#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2013-2014 NVIDIA Corporation
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------
#
#
#`define UVM_SEQ_ITEM_TASK_ERROR "Sequencer interface task not implemented"
#`define UVM_SEQ_ITEM_FUNCTION_ERROR "Sequencer interface function not implemented"
#
#//------------------------------------------------------------------------------
#//
#// CLASS: uvm_sqr_if_base #(REQ,RSP)
#//
#// This class defines an interface for sequence drivers to communicate with
#// sequencers. The driver requires the interface via a port, and the sequencer
#// implements it and provides it via an export.
#//
#//------------------------------------------------------------------------------
#
#virtual class uvm_sqr_if_base #(type T1=uvm_object, T2=T1);
#
#  // Task: get_next_item
#  //
#  // Retrieves the next available item from a sequence.  The call will block
#  // until an item is available.  The following steps occur on this call:
#  //
#  // 1 - Arbitrate among requesting, unlocked, relevant sequences - choose the
#  //     highest priority sequence based on the current sequencer arbitration
#  //     mode. If no sequence is available, wait for a requesting unlocked
#  //     relevant sequence,  then re-arbitrate.
#  // 2 - The chosen sequence will return from wait_for_grant
#  // 3 - The chosen sequence <uvm_sequence_base::pre_do> is called
#  // 4 - The chosen sequence item is randomized
#  // 5 - The chosen sequence <uvm_sequence_base::post_do> is called
#  // 6 - Return with a reference to the item
#  //
#  // Once <get_next_item> is called, <item_done> must be called to indicate the
#  // completion of the request to the sequencer.  This will remove the request
#  // item from the sequencer FIFO.
#
#  virtual task get_next_item(output T1 t);
#    uvm_report_error("get_next_item", `UVM_SEQ_ITEM_TASK_ERROR, UVM_NONE);
#  endtask
# 
#
#  // Task: try_next_item
#  //
#  // Retrieves the next available item from a sequence if one is available.
#  // Otherwise, the function returns immediately with request set to ~null~.
#  // The following steps occur on this call:
#  //
#  // 1 - Arbitrate among requesting, unlocked, relevant sequences - choose the
#  //     highest priority sequence based on the current sequencer arbitration
#  //     mode. If no sequence is available, return ~null~.
#  // 2 - The chosen sequence will return from wait_for_grant
#  // 3 - The chosen sequence <uvm_sequence_base::pre_do> is called
#  // 4 - The chosen sequence item is randomized
#  // 5 - The chosen sequence <uvm_sequence_base::post_do> is called
#  // 6 - Return with a reference to the item
#  //
#  // Once <try_next_item> is called, <item_done> must be called to indicate the
#  // completion of the request to the sequencer.  This will remove the request
#  // item from the sequencer FIFO.
#
#  virtual task try_next_item(output T1 t);
#    uvm_report_error("try_next_item", `UVM_SEQ_ITEM_TASK_ERROR, UVM_NONE);
#  endtask
#
#
#  // Function: item_done
#  //
#  // Indicates that the request is completed to the sequencer.  Any 
#  // <uvm_sequence_base::wait_for_item_done>
#  // calls made by a sequence for this item will return.
#  //
#  // The current item is removed from the sequencer FIFO.
#  //
#  // If a response item is provided, then it will be sent back to the requesting
#  // sequence. The response item must have its sequence ID and transaction ID
#  // set correctly, using the <uvm_sequence_item::set_id_info> method:
#  //
#  //|  rsp.set_id_info(req);
#  //
#  // Before <item_done> is called, any calls to peek will retrieve the current
#  // item that was obtained by <get_next_item>.  After <item_done> is called, peek
#  // will cause the sequencer to arbitrate for a new item.
#
#  virtual function void item_done(input T2 t = null);
#    uvm_report_error("item_done", `UVM_SEQ_ITEM_FUNCTION_ERROR, UVM_NONE);
#  endfunction
#
#  
#  // Task: wait_for_sequences
#  //
#  // Waits for a sequence to have a new item available. The default
#  // implementation in the sequencer calls
#  //  <uvm_wait_for_nba_region>.
#  // User-derived sequencers
#  // may override its <wait_for_sequences> implementation to perform some other
#  // application-specific implementation.
#
#  virtual task wait_for_sequences();
#    uvm_report_error("wait_for_sequences", `UVM_SEQ_ITEM_TASK_ERROR, UVM_NONE);
#  endtask
#
#
#  // Function: has_do_available
#  //
#  // Indicates whether a sequence item is available for immediate processing.
#  // Implementations should return 1 if an item is available, 0 otherwise. 
#
#  virtual function bit has_do_available();
#    uvm_report_error("has_do_available", `UVM_SEQ_ITEM_FUNCTION_ERROR, UVM_NONE);
#    return 0;
#  endfunction
#
#
#  //-----------------------
#  // uvm_tlm_blocking_slave_if
#  //-----------------------
#
#  // Task: get
#  //
#  // Retrieves the next available item from a sequence.  The call blocks until
#  // an item is available. The following steps occur on this call:
#  //
#  // 1 - Arbitrate among requesting, unlocked, relevant sequences - choose the
#  //     highest priority sequence based on the current sequencer arbitration
#  //     mode. If no sequence is available, wait for a requesting unlocked
#  //     relevant sequence, then re-arbitrate.
#  // 2 - The chosen sequence will return from <uvm_sequence_base::wait_for_grant>
#  // 3 - The chosen sequence <uvm_sequence_base::pre_do> is called
#  // 4 - The chosen sequence item is randomized
#  // 5 - The chosen sequence <uvm_sequence_base::post_do> is called
#  // 6 - Indicate <item_done> to the sequencer
#  // 7 - Return with a reference to the item
#  //
#  // When get is called, <item_done> may not be called.  A new item can be
#  // obtained by calling get again, or a response may be sent using either
#  // <put>, or uvm_driver::rsp_port.write().
#
#  virtual task get(output T1 t);
#    uvm_report_error("get", `UVM_SEQ_ITEM_TASK_ERROR, UVM_NONE);
#  endtask
#
#  // Task: peek
#  //
#  // Returns the current request item if one is in the sequencer FIFO.  If no
#  // item is in the FIFO, then the call will block until the sequencer has a new
#  // request. The following steps will occur if the sequencer FIFO is empty:
#  //
#  // 1 - Arbitrate among requesting, unlocked, relevant sequences - choose the
#  // highest priority sequence based on the current sequencer arbitration mode.
#  // If no sequence is available, wait for a requesting unlocked relevant
#  // sequence, then re-arbitrate.
#  //
#  // 2 - The chosen sequence will return from <uvm_sequence_base::wait_for_grant>
#  // 3 - The chosen sequence <uvm_sequence_base::pre_do> is called
#  // 4 - The chosen sequence item is randomized
#  // 5 - The chosen sequence <uvm_sequence_base::post_do> is called
#  //
#  // Once a request item has been retrieved and is in the sequencer FIFO,
#  // subsequent calls to peek will return the same item.  The item will stay in
#  // the FIFO until either get or <item_done> is called.
#
#  virtual task peek(output T1 t);
#    uvm_report_error("peek", `UVM_SEQ_ITEM_TASK_ERROR, UVM_NONE);
#  endtask
#
#
#  // Task: put
#  //
#  // Sends a response back to the sequence that issued the request. Before the
#  // response is put, it must have its sequence ID and transaction ID set to
#  // match the request.  This can be done using the
#  // <uvm_sequence_item::set_id_info> call:
#  //
#  //   rsp.set_id_info(req);
#  //
#  // While this is a task, it will not consume time (including delta cycles).
#  // The response will be put into the
#  // sequence response queue or it will be sent to the
#  // sequence response handler.
#  //
#   
#
#  virtual task put(input T2 t);
#    uvm_report_error("put", `UVM_SEQ_ITEM_TASK_ERROR, UVM_NONE);
#  endtask
#
#
#  // Function: put_response
#  //
#  // Sends a response back to the sequence that issued the request. Before the
#  // response is put, it must have its sequence ID and transaction ID set to
#  // match the request.  This can be done using the
#  // <uvm_sequence_item::set_id_info> call:
#  //
#  //   rsp.set_id_info(req);
#  //
#
#  virtual function void put_response(input T2 t);
#    uvm_report_error("put_response", `UVM_SEQ_ITEM_FUNCTION_ERROR, UVM_NONE);
#  endfunction
#
#  // Function: disable_auto_item_recording
#  //
#  // By default, item recording is performed automatically when
#  // get_next_item() and item_done() are called.
#  // However, this works only for simple, in-order, blocking transaction
#  // execution. For pipelined and out-of-order transaction execution, the
#  // driver must turn off this automatic recording and call
#  // <uvm_transaction::accept_tr>, <uvm_transaction::begin_tr>
#  // and <uvm_transaction::end_tr> explicitly at appropriate points in time.
#  //
#  // This methods be called at the beginning of the driver's ~run_phase()~ method.
#  // Once disabled, automatic recording cannot be re-enabled.
#  //
#  // For backward-compatibility, automatic item recording can be globally
#  // turned off at compile time by defining UVM_DISABLE_AUTO_ITEM_RECORDING
#
#  virtual function void disable_auto_item_recording();
#    uvm_report_error("disable_auto_item_recording", `UVM_SEQ_ITEM_FUNCTION_ERROR, UVM_NONE);
#  endfunction
#  
#  // Function: is_auto_item_recording_enabled
#  //
#  // Return TRUE if automatic item recording is enabled for this port instance.
#
#  virtual function bit is_auto_item_recording_enabled();
#    uvm_report_error("is_auto_item_recording_enabled", `UVM_SEQ_ITEM_FUNCTION_ERROR, UVM_NONE);
#    return 0;
#  endfunction
#endclass



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from uvm.base.uvm_component import UVMComponent
from uvm.macros import uvm_component_utils
#from uvm import *
from packet_pkg import packet


class gen(UVMComponent):

    def __init__(self, name, parent):
        super().__init__(name, parent)


    def get_packet(self):

        # use the factory to generate a packet
        pkt = packet.type_id.create("p", self)

        # randomize it
        pkt.randomize()  # cast to 'void' removed

        return pkt


#    //Use the macro in a class to implement factory registration along with other
#    //utilities (create, get_type_name). To do only factory registration, use
#    //the macro `uvm_component_utils(gen,"gen").
uvm_component_utils(gen)



#//
#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2013      NVIDIA Corporation
#//   Copyright 2013      Verilab, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------

# pytype: disable=attribute-error,not-writable

from typing import List, Optional

from .sv import sv
from .uvm_queue import UVMQueue
from .uvm_pool import UVMPool
from .uvm_globals import (uvm_report_error, uvm_report_fatal, uvm_report_warning,
    uvm_is_match, uvm_report_info)
from .uvm_object_globals import UVM_NONE, UVM_MEDIUM, UVM_HIGH
from ..macros import uvm_info
from ..uvm_macros import UVM_STRING_QUEUE_STREAMING_PACK



def m_has_wildcard(nm: str) -> bool:
    """ Returns True if given string has wildcard * or ? """
    for char in nm:
        if char == "*" or char == "?":
            return True
    return False


class uvm_factory_queue_class:
    def __init__(self):
        self.queue = UVMQueue()

#------------------------------------------------------------------------------
#
# CLASS- UVMFactoryOverride
#
# Internal class. Data structure for factory overrides to store all override
# related information.
#------------------------------------------------------------------------------

class UVMFactoryOverride:


    def __init__(self, full_inst_path="", orig_type_name="", orig_type=None,
            ovrd_type=None):
        if ovrd_type is None:
            uvm_report_fatal("NULLWR",
                "Attempting to register a null override object with the factory", UVM_NONE)
        self.full_inst_path = full_inst_path
        self.orig_type_name = orig_type_name
        if orig_type is not None:
            self.orig_type_name  = orig_type.get_type_name()

        self.orig_type      = orig_type
        self.ovrd_type_name = ovrd_type.get_type_name()
        self.ovrd_type      = ovrd_type
        self.selected = False
        self.used = 0



class UVMFactory:
    """
    UVMFactory is used to create objects of type `UVMComponent` and
    `UVMObject` (and their derived user-defined types).
    Object and component types are registered
    with the factory using lightweight proxies to the actual objects and
    components being created. The `UVMObjectRegistry` and
    `UVMComponentRegistry` class are used to proxy `UVMObject`
    and `UVMComponent`.

    The factory provides both name-based and type-based interfaces.

    type-based - The type-based interface is far less prone to errors in usage.

    name-based - The name-based interface is dominated
      by string arguments that can be misspelled and provided in the wrong order.
      Errors in name-based requests might only be caught at the time of the call,
      if at all.

    The `UVMFactory` is an abstract class. The UVM uses the `UVMDefaultFactory` class
    as its default factory implementation.

    See `UVMDefaultFactory` section for details on configuring and using the factory.
    """

    #// Group: Retrieving the factory

    @classmethod
    def get(cls) -> 'UVMFactory':
        """
        Function: get
        Static accessor for `UVMFactory`

        The static accessor is provided as a convenience wrapper
        around retrieving the factory via the `UVMCoreService.get_factory`
        method.

        .. code-block:: python

            # Using the uvm_coreservice_t:
            cs = UVMCoreService.get()
            f = cs.get_factory()

        .. code-block:: python

            # Not using the uvm_coreservice_t:
            f = UVMFactory.get()

        Returns:
            UVMFactory: Singleton instace of the factory
        """
        from .uvm_coreservice import UVMCoreService
        cs = UVMCoreService.get()
        return cs.get_factory()

    # NOTE: SV virtual functions not added from original uvm_factory, not needed in python

#------------------------------------------------------------------------------
#
# CLASS: uvm_default_factory
#
#------------------------------------------------------------------------------
#
# Default implementation of the UVM factory.


class UVMDefaultFactory(UVMFactory):

    m_debug_pass = False

    def __init__(self):
        self.m_override_info: List[UVMFactoryOverride] = []
        self.m_types = UVMPool()  # [uvm_object_wrapper] -> bit
        self.m_lookup_strs = UVMPool()
        self.m_type_names = UVMPool()
        self.m_type_overrides: List[UVMFactoryOverride] = []
        self.m_inst_override_queues = {}  # [uvm_object_wrapper] -> queue
        self.m_inst_override_name_queues = {}  # [string] -> queue
        self.m_wildcard_inst_overrides = UVMQueue()


    def register(self, obj) -> None:
        """
        Group: Registering Types

        Function: register

        Registers the given proxy object, `obj`, with the factory.
        Args:
            obj:
        """
        if obj is None:
            uvm_report_fatal("NULLWR",
                "Attempting to register a null object with the factory")

        if obj.get_type_name() != "" and obj.get_type_name() != "<unknown>":
            if self.m_type_names.exists(obj.get_type_name()):
                uvm_report_warning("TPRGED", ("Type name '" + obj.get_type_name()
                    + "' already registered with factory. No string-based lookup "
                    + "support for multiple types with the same type name."), UVM_NONE)
            else:
                self.m_type_names.add(obj.get_type_name(), obj)

        if self.m_types.exists(obj):
            if obj.get_type_name() != "" and obj.get_type_name() != "<unknown>":
                uvm_report_warning("TPRGED", ("Object type '" + obj.get_type_name()
                                 + "' already registered with factory. "), UVM_NONE)
        else:
            self.m_types.add(obj, 1)
            # If a named override happens before the type is registered, need to copy
            # the override queue.
            # Note:Registration occurs via static initialization, which occurs ahead of
            # procedural (e.g. initial) blocks. There should not be any preexisting overrides.
            tname = obj.get_type_name()
            if obj.get_type_name() in self.m_inst_override_name_queues:
                self.m_inst_override_queues[obj] = UVMQueue()
                self.m_inst_override_queues[obj] = self.m_inst_override_name_queues[tname]
                del self.m_inst_override_name_queues[tname]
            if self.m_wildcard_inst_overrides.size():
                if obj not in self.m_inst_override_queues:
                    self.m_inst_override_queues[obj] = UVMQueue()
                for i in range(0, self.m_wildcard_inst_overrides.size()):
                    wc_inst_override = self.m_wildcard_inst_overrides[i]
                    if uvm_is_match(wc_inst_override.orig_type_name, tname):
                        self.m_inst_override_queues[obj].push_back(wc_inst_override)

    def set_type_override_by_type(self, original_type, override_type,
            replace=True):
        """
        set_type_override_by_type
        -------------------------

        #def set_type_override_by_type (uvm_object_wrapper original_type,
                                                     uvm_object_wrapper override_type,
                                                     bit replace=1)
        Args:
            original_type:
            override_type:
            replace:
        """
        replaced = False

        # check that old and new are not the same
        if original_type == override_type:
            if original_type.get_type_name() == "" or original_type.get_type_name() == "<unknown>":
                uvm_report_warning("TYPDUP", "Original and override type "
                        + "arguments are identical", UVM_NONE)
            else:
                uvm_report_warning("TYPDUP", "Original and override type "
                        + "arguments are identical: "
                        + original_type.get_type_name(), UVM_NONE)

        # register the types if not already done so, for the benefit of string-based lookup
        if not self.m_types.exists(original_type):
            self.register(original_type)

        if not self.m_types.exists(override_type):
            self.register(override_type)

        # check for existing type override
        for index in range(len(self.m_type_overrides)):
            idx_orig_type = self.m_type_overrides[index].orig_type
            idx_type_name = self.m_type_overrides[index].orig_type_name
            if (idx_orig_type == original_type or
                    (idx_type_name != "<unknown>" and
                        idx_type_name != "" and
                        idx_type_name == original_type.get_type_name())):

                msg = ("Original object type '" + original_type.get_type_name()
                       + "' already registered to produce '"
                       + self.m_type_overrides[index].ovrd_type_name + "'")
                if replace is False:
                    msg = msg + ".  Set 'replace' argument to replace the existing entry."
                    uvm_report_info("TPREGD", msg, UVM_MEDIUM)
                    return

                msg = (msg + ".  Replacing with override to produce type '"
                    + override_type.get_type_name() + "'.")
                uvm_report_info("TPREGR", msg, UVM_MEDIUM)
                replaced = True
                self.m_type_overrides[index].orig_type = original_type
                self.m_type_overrides[index].orig_type_name = original_type.get_type_name()
                self.m_type_overrides[index].ovrd_type = override_type
                self.m_type_overrides[index].ovrd_type_name = override_type.get_type_name()



        #  make a new entry
        if replaced is False:
            # uvm_factory_override override
            override = UVMFactoryOverride(orig_type=original_type,
                           orig_type_name=original_type.get_type_name(),
                           full_inst_path="*",
                           ovrd_type=override_type)

            self.m_type_overrides.append(override)


    def set_type_override_by_name(self, original_type_name, override_type_name,
            replace=True):
        """
        set_type_override_by_name
        -------------------------
        Args:
            original_type_name:
            override_type_name:
            replace:
        """
        replaced = False
        original_type = None
        override_type = None

        if self.m_type_names.exists(original_type_name):
            original_type = self.m_type_names.get(original_type_name)

        if self.m_type_names.exists(override_type_name):
            override_type = self.m_type_names.get(override_type_name)

        # check that type is registered with the factory
        if override_type is None:
            uvm_report_error("TYPNTF", ("Cannot register override for original type '"
            + original_type_name + "' because the override type '"
            + override_type_name + "' is not registered with the factory."), UVM_NONE)
            return

        # check that old and new are not the same
        if original_type_name == override_type_name:
            uvm_report_warning("TYPDUP", ("Requested and actual type name "
                + " arguments are identical: " + original_type_name
                + ". Ignoring this override."), UVM_NONE)
            return

        for index in range(0, len(self.m_type_overrides)):
            ovrd_type_name = self.m_type_overrides[index].ovrd_type_name
            if self.m_type_overrides[index].orig_type_name == original_type_name:
                if not replace:
                    uvm_report_info("TPREGD", ("Original type '" + original_type_name
                      + "' already registered to produce '" + ovrd_type_name
                      + "'.  Set 'replace' argument to replace the existing entry."), UVM_MEDIUM)
                    return

                uvm_report_info("TPREGR", ("Original object type '" + original_type_name
                  + "' already registered to produce '" + ovrd_type_name
                  + "'.  Replacing with override to produce type '"
                  + override_type_name + "'."), UVM_MEDIUM)
                replaced = True
                self.m_type_overrides[index].ovrd_type = override_type
                self.m_type_overrides[index].ovrd_type_name = override_type_name

        if original_type is None:
            self.m_lookup_strs[original_type_name] = 1

        if not replaced:
            # uvm_factory_override override
            override = UVMFactoryOverride(orig_type=original_type,
                           orig_type_name=original_type_name,
                           full_inst_path="*",
                           ovrd_type=override_type)

            self.m_type_overrides.append(override)
            self.m_type_names.add(original_type_name, override.ovrd_type)

    def check_inst_override_exists(self, original_type, override_type,
            full_inst_path):
        """
        check_inst_override_exists
        --------------------------
        #def check_inst_override_exists (uvm_object_wrapper original_type,
                                             uvm_object_wrapper override_type,
                                             string full_inst_path)
        Args:
            original_type:
            override_type:
            full_inst_path:
        Returns:
        """
        override = None  # uvm_factory_override
        qc = None  # uvm_factory_queue_class

        if original_type in self.m_inst_override_queues:
            qc = self.m_inst_override_queues[original_type]
        else:
            return 0

        for index in range(len(qc.queue)):
            override = qc.queue[index]
            if (override.full_inst_path == full_inst_path and
                    override.orig_type == original_type and
                    override.ovrd_type == override_type and
                    override.orig_type_name == original_type.get_type_name()):
                uvm_report_info("DUPOVRD", "Instance override for '"
                    + original_type.get_type_name() + "' already exists: override type '"
                    + override_type.get_type_name() + "' with full_inst_path '"
                    + full_inst_path + "'", UVM_HIGH)
                return 1
        return 0


    def set_inst_override_by_type(self, original_type, override_type,
            full_inst_path):
        """
        set_inst_override_by_type
        -------------------------

        #def set_inst_override_by_type (uvm_object_wrapper original_type,
                                                             uvm_object_wrapper override_type,
                                                             string full_inst_path)
        Args:
            original_type:
            override_type:
            full_inst_path:
        """

        override = None  # uvm_factory_override

        # register the types if not already done so
        if not self.m_types.exists(original_type):
            self.register(original_type)

        if not self.m_types.exists(override_type):
            self.register(override_type)

        if self.check_inst_override_exists(original_type,override_type,full_inst_path):
            return

        if original_type not in self.m_inst_override_queues:
            self.m_inst_override_queues[original_type] = UVMQueue()

        override = UVMFactoryOverride(full_inst_path=full_inst_path,
                orig_type=original_type,
                orig_type_name=original_type.get_type_name(),
                ovrd_type=override_type)


        self.m_inst_override_queues[original_type].push_back(override)


    # set_inst_override_by_name
    # -------------------------
    # TODO
    #function void uvm_default_factory::set_inst_override_by_name (
    # string original_type_name,
    #                                                      string override_type_name,
    #                                                      string full_inst_path)
    #
    #  uvm_factory_override override
    #  uvm_object_wrapper original_type
    #  uvm_object_wrapper override_type
    #
    #  if(self.m_type_names.exists(original_type_name))
    #    original_type = self.m_type_names[original_type_name]
    #
    #  if(self.m_type_names.exists(override_type_name))
    #    override_type = self.m_type_names[override_type_name]
    #
    #  // check that type is registered with the factory
    #  if (override_type  is None):
    #    uvm_report_error("TYPNTF", {"Cannot register instance override with type name '",
    #    original_type_name,"' and instance path '",full_inst_path,"' because the type it's supposed ",
    #    "to produce, '",override_type_name,"', is not registered with the factory."}, UVM_NONE)
    #    return
    #  end
    #
    #  if (original_type  is None)
    #      m_lookup_strs[original_type_name] = 1
    #
    #  override = new(.full_inst_path(full_inst_path),
    #                 .orig_type(original_type),
    #                 .orig_type_name(original_type_name),
    #                 .ovrd_type(override_type))
    #
    #  if(original_type != null):
    #    if (check_inst_override_exists(original_type,override_type,full_inst_path))
    #      return
    #    if(!m_inst_override_queues.exists(original_type))
    #      m_inst_override_queues[original_type] = new
    #    m_inst_override_queues[original_type].queue.push_back(override)
    #  end
    #  else:
    #    if(m_has_wildcard(original_type_name)):
    #       foreach(self.m_type_names[i]):
    #         if(uvm_is_match(original_type_name,i)):
    #           this.set_inst_override_by_name(i, override_type_name, full_inst_path)
    #         end
    #       end
    #       m_wildcard_inst_overrides.push_back(override)
    #    end
    #    else:
    #      if(!self.m_inst_override_name_queues.exists(original_type_name))
    #        self.m_inst_override_name_queues[original_type_name] = new
    #      self.m_inst_override_name_queues[original_type_name].queue.push_back(override)
    #    end
    #  end
    #
    #endfunction

    def create_object_by_name(self, requested_type_name, parent_inst_path="",
              name=""):
        """
        create_object_by_name
        ---------------------
        Args:
            requested_type_name:
            parent_inst_path:
            name:
        Returns:
        """
        inst_path = self._get_inst_path(parent_inst_path, name)

        self.m_override_info.clear()
        wrapper = self.find_override_by_name(requested_type_name, inst_path)

        # if no override exists, try to use requested_type_name directly
        if wrapper is None:
            if not self.m_type_names.exists(requested_type_name):
                uvm_report_warning("BDTYP", ("Cannot create an object of type '"
                    + requested_type_name + "' because it is not registered with the factory."),
                    UVM_NONE)
                return None
            wrapper = self.m_type_names.get(requested_type_name)
        return wrapper.create_object(name)

    def create_object_by_type(self, requested_type, parent_inst_path="", name=""):
        """
        create_object_by_type
        ---------------------
        Args:
            requested_type:
            parent_inst_path:
            name:
        Returns:
        """
        if requested_type is None:
            uvm_report_fatal("REQ_TYPE_NONE", "Requested type object was None")
        full_inst_path = self._get_inst_path(parent_inst_path, name)

        self.m_override_info.clear()
        requested_type = self.find_override_by_type(requested_type, full_inst_path)
        if requested_type is None:
            uvm_report_fatal("REQ_TYPE_NONE", "Requested type object was None after override")
        return requested_type.create_object(name)

    def _get_inst_path(self, parent_inst_path, name):
        inst_path = ""
        if parent_inst_path == "":
            inst_path = name
        elif name != "":
            inst_path = parent_inst_path + "." + name
        else:
            inst_path = parent_inst_path
        return inst_path

    def create_component_by_name(self, requested_type_name,
            parent_inst_path, name, parent):
        """
        create_component_by_name
        ------------------------
        Args:
            requested_type_name:
            parent_inst_path:
            name:
            parent:
        Returns:
        """

        inst_path = self._get_inst_path(parent_inst_path, name)

        self.m_override_info.clear()
        wrapper = self.find_override_by_name(requested_type_name, inst_path)

        # if no override exists, try to use requested_type_name directly
        if wrapper is None:
            if not self.m_type_names.exists(requested_type_name):
                uvm_report_warning("BDTYP", ("Cannot create a component of type '"
                    + requested_type_name + "' because it is not registered with the factory."),
                    UVM_NONE)
                return None
            wrapper = self.m_type_names.get(requested_type_name)

        return wrapper.create_component(name, parent)

    def create_component_by_type(self, requested_type, parent_inst_path, name,
            parent):
        """
        create_component_by_type
        ------------------------
        Args:
            requested_type:
            parent_inst_path:
            name:
            parent:
        Returns:
        """
        full_inst_path = self._get_inst_path(parent_inst_path, name)

        self.m_override_info.clear()
        requested_type = self.find_override_by_type(requested_type, full_inst_path)
        return requested_type.create_component(name, parent)

    # find_wrapper_by_name
    # ------------
    # TODO
    def find_wrapper_by_name(self, type_name: str) -> Optional['UVMObjectWrapper']:
        if self.m_type_names.exists(type_name):
            return self.m_type_names[type_name]
        uvm_report_warning("UnknownTypeName", ("find_wrapper_by_name: Type name '"
            + type_name + "' not registered with the factory."), UVM_NONE)

    def find_override_by_name(
            self, requested_type_name, full_inst_path) -> Optional['UVMObjectWrapper']:
        """
        find_override_by_name
        ---------------------
        Args:
            requested_type_name (str):
            full_inst_path (str):
        Returns:
            UVMObjectWrapper
        """
        rtype = None
        qc = UVMQueue()
        lindex = None
        override = None

        if self.m_type_names.exists(requested_type_name):
            rtype = self.m_type_names.get(requested_type_name)

        if full_inst_path != "":
            if rtype is None:
                if requested_type_name in self.m_inst_override_name_queues:
                    qc = self.m_inst_override_name_queues[requested_type_name]
            else:
                if rtype in self.m_inst_override_queues:
                    qc = self.m_inst_override_queues[rtype]

            if qc is not None:
                for index in range(0, qc.size()):
                    match_ok = uvm_is_match(qc[index].orig_type_name, requested_type_name)
                    if match_ok and uvm_is_match(qc[index].full_inst_path, full_inst_path):
                        self.m_override_info.append(qc[index])
                        if UVMDefaultFactory.m_debug_pass:
                            if override is None:
                                override = qc[index].ovrd_type
                                qc[index].selected = 1
                                lindex = qc[index]
                        else:
                            qc[index].used += 1
                            if qc[index].ovrd_type.get_type_name() == requested_type_name:
                                return qc[index].ovrd_type
                            else:
                                return self.find_override_by_type(qc[index].ovrd_type,full_inst_path)

        ovrd_ok = rtype not in self.m_inst_override_queues and self.m_wildcard_inst_overrides.size() > 0
        if rtype is not None and ovrd_ok:
            self.m_inst_override_queues[rtype] = UVMQueue()
            for i in range(0, self.m_wildcard_inst_overrides.size()):
                if uvm_is_match(self.m_wildcard_inst_overrides.get(i).orig_type_name,
                        requested_type_name):
                    self.m_inst_override_queues[rtype].push_back(self.m_wildcard_inst_overrides.get(i))

        # type override - exact match
        for index in range(0, len(self.m_type_overrides)):
            if self.m_type_overrides[index].orig_type_name == requested_type_name:
                self.m_override_info.append(self.m_type_overrides[index])
                if UVMDefaultFactory.m_debug_pass:
                    if override is None:
                        override = self.m_type_overrides[index].ovrd_type
                        self.m_type_overrides[index].selected = 1
                        lindex = self.m_type_overrides[index]
                else:
                    self.m_type_overrides[index].used += 1
                    return self.find_override_by_type(self.m_type_overrides[index].ovrd_type,full_inst_path)

        if UVMDefaultFactory.m_debug_pass and override is not None:
            lindex.used += 1
            return self.find_override_by_type(override, full_inst_path)

        # No override found
        return None

    def find_override_by_type(self, requested_type, full_inst_path):
        """
        find_override_by_type
        ---------------------
        Args:
            requested_type:
            full_inst_path:
        Returns:
        """
        override = None
        lindex = None
        qc = None
        if requested_type in self.m_inst_override_queues:
            qc = self.m_inst_override_queues[requested_type]

        for index in range(0, len(self.m_override_info)):
            if self.m_override_info[index].orig_type == requested_type:
                uvm_report_error("OVRDLOOP", "Recursive loop detected while finding override.", UVM_NONE)
                if UVMDefaultFactory.m_debug_pass is False:
                    self.debug_create_by_type(requested_type, full_inst_path)

                self.m_override_info[index].used += 1
                return requested_type

        # inst override; return first match; takes precedence over type overrides
        if full_inst_path != "" and qc is not None:
            for index in range(0, qc.size()):
                if self.are_args_ok(qc[index], requested_type, full_inst_path):
                    self.m_override_info.append(qc[index])
                    if UVMDefaultFactory.m_debug_pass:
                        if override is None:
                            override = qc[index].ovrd_type
                            qc[index].selected = 1
                            lindex = qc[index]
                    else:
                        qc[index].used += 1
                        if qc[index].ovrd_type == requested_type:
                            return requested_type
                        else:
                            return self.find_override_by_type(qc[index].ovrd_type,full_inst_path)

        # type override - exact match
        for index in range(0, len(self.m_type_overrides)):
            if self.args_are_ok_again(self.m_type_overrides[index],requested_type):
                self.m_override_info.append(self.m_type_overrides[index])
                if UVMDefaultFactory.m_debug_pass:
                    if override is None:
                        override = self.m_type_overrides[index].ovrd_type
                        self.m_type_overrides[index].selected = 1
                        lindex = self.m_type_overrides[index]
                else:
                    self.m_type_overrides[index].used += 1
                    if self.m_type_overrides[index].ovrd_type == requested_type:
                        return requested_type
                    else:
                        return self.find_override_by_type(self.m_type_overrides[index].ovrd_type,
                            full_inst_path)

        if UVMDefaultFactory.m_debug_pass and override is not None:
            lindex.used += 1
            if override == requested_type:
                return requested_type
            else:
                return self.find_override_by_type(override,full_inst_path)

        return requested_type
        #endfunction

    def convert2string(self, all_types=True) -> str:
        """
        tpoikela: Added this to access factory string repr without print()

        Args:
            all_types (bool): If True, adds all possible types
        Returns:
            str: Factory converted to string
        """
        key = ""
        sorted_override_queues = {}  # uvm_factory_queue_class s[string]
        qs = UVMQueue()
        tmp = ""
        id = 0
        obj = None  # uvm_object_wrapper obj

        # sort the override queues
        for key in self.m_inst_override_queues:
            obj = self.m_inst_override_queues[key]
            tmp = obj.get_type_name()
            if tmp == "":
                tmp = "__unnamed_id_" + str(id)
                id += 1
            sorted_override_queues[tmp] = self.m_inst_override_queues[key]

        for key in self.m_inst_override_name_queues:
            sorted_override_queues[key] = self.m_inst_override_name_queues[key]

        qs.push_back("\n#### Factory Configuration (*)\n\n")

        # print instance overrides
        if len(self.m_type_overrides) == 0 and len(sorted_override_queues) == 0:
            qs.push_back("  No instance or type overrides are registered with this factory\n")
        else:
            max1 = 0
            max2 = 0
            max3 = 0
            dash = "-" * 100
            space = " " * 100
            # print instance overrides
            if len(sorted_override_queues) == 0:
                qs.push_back("No instance overrides are registered with this factory\n")
            else:
                for key, qc in sorted_override_queues.items():
                    for i in range(len(qc.queue)):
                        orig_type_name = qc.queue[i].orig_type_name
                        full_inst_path = qc.queue[i].full_inst_path
                        ovrd_type_name = qc.queue[i].ovrd_type_name
                        if len(orig_type_name) > max1:
                            max1 = len(orig_type_name)
                        if len(full_inst_path) > max2:
                            max2 = len(full_inst_path)
                        if len(ovrd_type_name) > max3:
                            max3 = len(ovrd_type_name)

                if max1 < 14:
                    max1 = 14
                if max2 < 13:
                    max2 = 13
                if max3 < 13:
                    max3 = 13

                qs.push_back("Instance Overrides:\n\n")
                qs.push_back(sv.sformatf("  %0s%0s  %0s%0s  %0s%0s\n","Requested Type",
                    space[1:max1-14], "Override Path", space[1:max2-13],
                    "Override Type", space[1:max3-13]))
                qs.push_back(sv.sformatf("  %0s  %0s  %0s\n",dash[1:max1],
                    dash[1:max2], dash[1:max3]))

                for key, qc in sorted_override_queues.items():
                    # qc = sorted_override_queues[j]
                    for i in range(len(qc.queue)):
                        qs.push_back(sv.sformatf("  %0s%0s  %0s%0s",qc.queue[i].orig_type_name,
                            space[1:max1-len(qc.queue[i].orig_type_name)],
                            qc.queue[i].full_inst_path,
                            space[1:max2-len(qc.queue[i].full_inst_path)]))
                        qs.push_back(sv.sformatf("  %0s\n", qc.queue[i].ovrd_type_name))

            # print type overrides
            if len(self.m_type_overrides) == 0:
                qs.push_back("\nNo type overrides are registered with this factory\n")
            else:
                # Resize for type overrides
                if max1 < 14:
                    max1 = 14
                if max2 < 13:
                    max2 = 13
                if max3 < 13:
                    max3 = 13

                for i in range(len(self.m_type_overrides)):
                    if len(self.m_type_overrides[i].orig_type_name) > max1:
                        max1 = len(self.m_type_overrides[i].orig_type_name)
                    if len(self.m_type_overrides[i].ovrd_type_name) > max2:
                        max2 = len(self.m_type_overrides[i].ovrd_type_name)
                if max1 < 14:
                    max1 = 14
                if max2 < 13:
                    max2 = 13
                qs.push_back("\nType Overrides:\n\n")
                qs.push_back(sv.sformatf("  %0s%0s  %0s%0s\n","Requested Type", space[1:max1-14],
                    "Override Type", space[1:max2-13]))
                qs.push_back(sv.sformatf("  %0s  %0s\n",dash[1:max1],
                    dash[1:max2]))
                for index in range(len(self.m_type_overrides)):
                    qs.push_back(sv.sformatf("  %0s%0s  %0s\n",
                             self.m_type_overrides[index].orig_type_name,
                             space[1:max1-len(self.m_type_overrides[index].orig_type_name)],
                             self.m_type_overrides[index].ovrd_type_name))

        # print all registered types, if all_types >= 1
        if all_types >= 1 and self.m_type_names.has_first():
            banner = False
            qs.push_back(sv.sformatf("\nAll types registered with the factory: %0d total\n",self.m_types.num()))
            key = self.m_type_names.first()
            while key is not None:
                # filter out uvm_ classes (if all_types<2) and non-types (lookup strings)
                if (not (all_types < 2 and uvm_is_match("uvm_*",
                       self.m_type_names[key].get_type_name())) and
                       key == self.m_type_names[key].get_type_name()):
                    if not banner:
                        qs.push_back("  Type Name\n")
                        qs.push_back("  ---------\n")
                        banner = True
                    qs.push_back(sv.sformatf("  %s\n", self.m_type_names[key].get_type_name()))
                if self.m_type_names.has_next():
                    key = self.m_type_names.next()
                else:
                    break
            # end while(self.m_type_names.next(key))

        qs.push_back("(*) Types with no associated type name will be printed as <unknown>\n\n####\n\n")
        return UVM_STRING_QUEUE_STREAMING_PACK(qs)

    def print_factory(self, all_types=True):
        """
        print
        -----
        Args:
            all_types:
        """
        fact_str = self.convert2string(all_types)
        uvm_info("UVM/FACTORY/PRINT", fact_str, UVM_NONE)

    # debug_create_by_name
    # --------------------
    # TODO
    def debug_create_by_name(self, requested_type_name, parent_inst_path="", name=""):
        self.m_debug_create(requested_type_name, None, parent_inst_path, name)

    # debug_create_by_type
    # --------------------

    def debug_create_by_type(self, requested_type, parent_inst_path="", name=""):
        self.m_debug_create("", requested_type, parent_inst_path, name)

    # m_debug_create
    # --------------
    # TODO
    def m_debug_create(self, requested_type_name, requested_type,
            parent_inst_path, name):
        full_inst_path = self._get_inst_path(parent_inst_path, name)
        result = None
        self.m_override_info.clear()

        if requested_type is None:
            if (not self.m_type_names.exists(requested_type_name) and
                    not self.m_lookup_strs.exists(requested_type_name)):
                uvm_report_warning("Factory Warning", {"The factory does not recognize '",
                    requested_type_name,"' as a registered type."}, UVM_NONE)
                return
            UVMDefaultFactory.m_debug_pass = 1
            result = self.find_override_by_name(requested_type_name,full_inst_path)
        else:
            UVMDefaultFactory.m_debug_pass = 1
            if not self.m_types.exists(requested_type):
                self.register(requested_type)
            result = self.find_override_by_type(requested_type,full_inst_path)
            if requested_type_name == "":
                requested_type_name = requested_type.get_type_name()
        self.m_debug_display(requested_type_name, result, full_inst_path)
        UVMDefaultFactory.m_debug_pass = 0
        for obj in self.m_override_info:
            obj.selected = 0

    # m_debug_display
    # ---------------
    # TODO
    def m_debug_display(self, requested_type_name, result, full_inst_path):
        pass
        max1 = 0
        max2 = 0
        max3 = 0
        dash = 100 * "-"
        space = 100 * " "
        qs = []

        qs.append("\n#### Factory Override Information (*)\n\n")
        qs.append(sv.sformatf("Given a request for an object of type '%s' with "
            + "an instance\npath of '%s' the factory encountered\n\n",
            requested_type_name,full_inst_path))

        if len(self.m_override_info) == 0:
            qs.append("no relevant overrides.\n\n")
        else:
            qs.append("the following relevant overrides. "
                + "An 'x' next to a match indicates a\nmatch that was ignored.\n\n")

            for i in range(len(self.m_override_info)):
                if (len(self.m_override_info[i].orig_type_name) > max1):
                    max1 = len(self.m_override_info[i].orig_type_name)
                if (len(self.m_override_info[i].full_inst_path) > max2):
                    max2 = len(self.m_override_info[i].full_inst_path)
                if (len(self.m_override_info[i].ovrd_type_name) > max3):
                    max3 = len(self.m_override_info[i].ovrd_type_name)

            if max1 < 13:
                max1 = 13
            if max2 < 13:
                max2 = 13
            if max3 < 13:
                max3 = 13

            qs.append(sv.sformatf("Original Type%0s  Instance Path%0s  Override Type%0s\n",
                space[1:max1-12],space[1:max2-12],space[1:max3-12]))
            #space.substr(1,max1-13),space.substr(1,max2-13),space.substr(1,max3-13)))

            qs.append(sv.sformatf("  %0s  %0s  %0s\n",
                dash[1:max1], dash[1:max2], dash[1:max3]))

            for i in range(len(self.m_override_info)):
                orig_type_len = len(self.m_override_info[i].orig_type_name)
                full_inst_path_len = len(self.m_override_info[i].full_inst_path)
                ovrd_type_len = len(self.m_override_info[i].ovrd_type_name)
                is_sel_str = "x "
                if self.m_override_info[i].selected:
                    is_sel_str = "  "

                qs.append(sv.sformatf("%s%0s%0s",
                    is_sel_str,
                    self.m_override_info[i].orig_type_name,
                    space[1:max1-orig_type_len]))
                qs.append(sv.sformatf("  %0s%0s", self.m_override_info[i].full_inst_path,
                    space[1:max2-full_inst_path_len]))
                qs.append(sv.sformatf("  %0s%0s", self.m_override_info[i].ovrd_type_name,
                    space[1:max3-ovrd_type_len]))
                if self.m_override_info[i].full_inst_path == "*":
                    qs.append("  <type override>")
                else:
                    qs.append("\n")
            qs.append("\n")

        qs.append("Result:\n\n")
        chosen_type_name = requested_type_name
        if result is not None:
            chosen_type_name = result.get_type_name()
        qs.append(sv.sformatf("  The factory will produce an object of type '%0s'\n",
            chosen_type_name))

        qs.append("\n(*) Types with no associated type name will be printed as <unknown>\n\n####\n\n")
        uvm_info("UVM/FACTORY/DUMP", UVM_STRING_QUEUE_STREAMING_PACK(qs),UVM_NONE)

    # Internal helper functions

    def are_args_ok(self, qc_elem, requested_type, full_inst_path) -> bool:
        name_ok = (qc_elem.orig_type == requested_type or
                   (qc_elem.orig_type_name != "<unknown>" and
                    qc_elem.orig_type_name != "" and
                    qc_elem.orig_type_name == requested_type.get_type_name()))
        return name_ok and uvm_is_match(qc_elem.full_inst_path, full_inst_path)

    def args_are_ok_again(self, type_override, requested_type) -> bool:
        match_ok = (type_override.orig_type_name != "<unknown>" and
             type_override.orig_type_name != "" and
             requested_type is not None and
             type_override.orig_type_name == requested_type.get_type_name())
        return type_override.orig_type == requested_type or match_ok



class UVMObjectWrapper:
    """
    The UVMObjectWrapper provides an abstract interface for creating object and
    component proxies. Instances of these lightweight proxies, representing every
    `UVMObject`-based and `UVMComponent`-based object available in the test
    environment, are registered with the `UVMFactory`. When the factory is
    called upon to create an object or component, it finds and delegates the
    request to the appropriate proxy.
    """

    def create_object(self, name=""):
        """
        Creates a new object with the optional `name`.
        An object proxy (e.g., <uvm_object_registry #(T,Tname)>) implements this
        method to create an object of a specific type, T.

        Args:
            name (str):
        Returns:
            UVMObject|None
        """
        return None

    def create_component(self, name, parent):
        """
        Function: create_component

        Creates a new component, passing to its constructor the given `name` and
        `parent`. A component proxy (e.g. <uvm_component_registry #(T,Tname)>)
        implements this method to create a component of a specific type, T.

        Args:
            name (str):
            parent (UVMComponent): Parent of the created component.
        Returns:
            UVMComponent|None
        """
        return None

    def get_type_name(self):
        """
        Derived classes implement this method to return the type name of the object
        created by `create_component` or `create_object`. The factory uses this
        name when matching against the requested type in name-based lookups.

        Returns:
            str: Name of the type
        """
        return "<unknown_type>"

# pytype: enable=attribute-error,not-writable



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010-2011 Mentor Graphics Corporation
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from cocotb.triggers import *

from uvm.base.uvm_callback import *
from uvm.comps.uvm_monitor import UVMMonitor
from uvm.tlm1 import *
from uvm.macros import *

from .apb_rw import *


class apb_monitor_cbs(UVMCallback):
    def trans_observed(self, xactor, cycle):
        pass


class apb_monitor(UVMMonitor):

    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.ap = UVMAnalysisPort("ap", self)
        self.sigs = None  # passive
        self.cfg = None  # apb_config
        self.errors = 0
        self.num_items = 0
        self.tag = "APB_MONITOR"


    def build_phase(self, phase):
        super().build_phase(phase)
        agent = self.get_parent()
        if agent is not None:
            self.sigs = agent.vif
        else:
            arr = []
            if UVMConfigDb.get(self, "", "vif", arr):
                uvm_info("APB/MON/NOVIF", "Got vif through ConfigDb for APB monitor instance")
                self.sigs = arr[0]
        if self.sigs is None:
            uvm_fatal("APB/MON/NOVIF", "No virtual interface specified for self monitor instance")


    
    async def run_phase(self, phase):
        while True:
            tr = None

            # Wait for a SETUP cycle
            while True:
                await self.sample_delay()
                if (self.sigs.psel == 1 and
                   self.sigs.penable == 0):
                    break

            tr = apb_rw.type_id.create("tr", self)

            tr.kind = apb_rw.READ
            if int(self.sigs.pwrite) == 1:
                tr.kind = apb_rw.WRITE
            tr.addr = self.sigs.paddr.value.integer

            await self.sample_delay()
            if int(self.sigs.penable) != 1:
                val = int(self.sigs.penable)
                self.errors += 1
                uvm_error("APB", "APB protocol violation: SETUP cycle not followed by ENABLE cycle"
                        + " |Exp: 1, got: " + str(val))

            if tr.kind == apb_rw.READ:
                tr.data = self.sigs.prdata.value.integer
            else:
                tr.data = self.sigs.pwdata.value.integer

            self.trans_observed(tr)
            #TODO uvm_do_callbacks(apb_monitor,apb_monitor_cbs, self.trans_observed(self, tr))
            self.num_items += 1
            self.ap.write(tr)
            uvm_info(self.tag, "Sampled APB item: " + tr.convert2string(),
                UVM_HIGH)

    def trans_observed(self, tr):
        pass


    
    async def sample_delay(self):
        await RisingEdge(self.sigs.clk)
        await Timer(1, "NS")

uvm_component_utils(apb_monitor)




import cocotb
import re


class uvm_hdl():

    dut = None

    re_brackets = re.compile(r'(\w+)\[(\d+)\]')

    @classmethod
    def set_dut(cls, dut):
        cls.dut = dut
        cls.SIM_NAME = cocotb.SIM_NAME

    @classmethod
    def split_hdl_path(cls, path):
        if cls.SIM_NAME == 'Verilator':
            res = []
            spl = path.split('.')
            for name in spl:
                match = cls.re_brackets.search(name)
                if match is not None:
                    res.append(match[1])
                    # Throw exception if Conversion fails
                    res.append(int(match[2]))
                else:
                    res.append(name)
            return res
        else:
            return path.split('.')

    @classmethod
    def uvm_hdl_read(cls, path, value):
        if cls.dut is not None:
            curr_obj = cls.dut
            #split = path.split('.')
            split = cls.split_hdl_path(path)
            for spl in split:
                curr_obj = cls.get_next_obj(curr_obj, spl)
            if curr_obj is not None:
                value.append(int(curr_obj))
                return 1
            return 0
        else:
            raise Exception("dut is None. Use uvm_hdl.set_dut(dut) in your @cocotb.test()")


    @classmethod
    def uvm_hdl_deposit(cls, path, value):
        if cls.dut is not None:
            curr_obj = cls.dut
            #split = path.split('.')
            split = cls.split_hdl_path(path)
            for spl in split:
                curr_obj = cls.get_next_obj(curr_obj, spl)
                #if hasattr(curr_obj, spl):
                #    curr_obj = getattr(curr_obj, spl)
                #else:
                #    continue
            if curr_obj is not None:
                curr_obj.value = value
                return 1
            return 0
        else:
            raise Exception("dut is None. Use uvm_hdl.set_dut(dut) in your @cocotb.test()")


    @classmethod
    def get_next_obj(cls, curr_obj, spl):
        if isinstance(spl, int):
            #return curr_obj[spl]
            return curr_obj[spl]
        elif hasattr(curr_obj, spl):
            curr_obj = getattr(curr_obj, spl)
        return curr_obj



#
#------------------------------------------------------------------------------
#   Copyright 2011 Mentor Graphics Corporation
#   Copyright 2011 Cadence Design Systems, Inc.
#   Copyright 2011 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http:#www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------

from typing import List
from ..macros.uvm_message_defines import uvm_info
from .uvm_object_globals import UVM_NONE


class UVMSpellChkr:

    max_val = 1

    #typedef T tab_t[string];
    #static const int unsigned max = '1;

    @classmethod
    def check(cls, strtab, s):
        """
        Primary interface to the spell checker.  The function takes two
        arguments, a table of strings and a string to check.  The table is
        organized as an associative array of type T.  E.g.

           T strtab[string]

        It does not matter what T is since we are only concerned with the
        string keys. However, we need T in order to make argument types
        match.

        First, we do the simple thing and see if the string already is in
        the string table by calling the ~exists()~ method.  If it does exist
        then there is a match and we're done.  If the string doesn't exist
        in the table then we invoke the spell checker algorithm to see if
        our string is a misspelled variation on a string that does exist in
        the table.

        The main loop traverses the string table computing the levenshtein
        distance between each string and the string we are checking.  The
        strings in the table with the minimum distance are considered
        possible alternatives.  There may be more than one string in the
        table with a minimum distance. So all the alternatives are stored
        in a queue.

        Note: This is not a particularly efficient algorithm.  It requires
        computing the levenshtein distance for every string in the string
        table.  If that list were very large the run time could be long.
        For the resources application in UVM probably the size of the
        string table is not excessive and run times will be fast enough.
        If, on average, that proves to be an invalid assumption then we'll
        have to find ways to optimize this algorithm.

        note: strtab should not be modified inside check()

        Args:
            strtab:
            s:
        Returns:
            bool: True if check matches, False otherwise
        """
        key = ""
        distance = 0
        min_val = 0
        min_key = []

        if s in strtab:
            return True

        min_val = UVMSpellChkr.max_val
        for key in strtab:
            distance = UVMSpellChkr.levenshtein_distance(key, s)

            # A distance < 0 means either key, s, or both are empty.  This
            # should never happen here but we check for that condition just
            # in case.
            if distance < 0:
                continue

            if distance < min_val:
                # set a new minimum.  Clean out the queue since previous
                # alternatives are now invalidated.
                min_val = distance
                min_key = []
                min_key.append(key)
                continue

            if distance == min_val:
                min_key.append(key)

        # if (min == max) then the string table is empty
        if min_val == UVMSpellChkr.max_val:
            uvm_info("UVM/CONFIGDB/SPELLCHK",
                "{} not located, no alternatives to suggest".format(s), UVM_NONE)
        else:
            # dump all the alternatives with the minimum distance
            q = []
            for i in range(0, len(min_key)):
                q.append(min_key[i])
                q.append("|")
                if len(q) > 0:
                    q.pop()
                    uvm_info("UVM/CONFIGDB/SPELLCHK",
                            "{} not located, did you mean {}".format(s,
                                "".join(q)),UVM_NONE)
        return False

    @classmethod
    def levenshtein_distance(cls, s, t) -> int:
        """
        levenshtein_distance

        Compute levenshtein distance between s and t
        The Levenshtein distance is defined as The smallest number of
        insertions, deletions, and substitutions required to change one
        string into another.  There is a tremendous amount of information
        available on Levenshtein distance on the internet.  Two good
        sources are wikipedia and nist.gov.  A nice, simple explanation of
        the algorithm is at
        http:#www.codeproject.com/KB/recipes/Levenshtein.aspx.  Use google
        to find others.

        This implementation of the Levenshtein
        distance computation algorithm is a SystemVerilog adaptation of the
        C implementatiion located at http:#www.merriampark.com/ldc.htm.

        Args:
            s:
            t:
        Returns:
        """
        k        = 0
        i        = 0
        j        = 0
        n        = 0
        m        = 0
        cost     = 0
        distance = 0
        d: List[int] = []

        #Step 1
        n = len(s) + 1
        m = len(t) + 1

        if n == 1 or m == 1:
            return -1  # negative return value means that one or both strings are empty.

        d = m*n * [0]

        #Step 2
        for k in range(0, n):
            d[k] = k

        for k in range(0, m):
            d[k*n] = k

        #Steps 3 and 4
        for i in range(1, n):
            for j in range(1, m):
                #Step 5
                cost = s[i-1] != t[j-1]

                #Step 6
                d[j*n+i] = UVMSpellChkr.minimum(d[(j-1)*n+i]+1, d[j*n+i-1]+1, d[(j-1)*n+i-1]+cost)

        distance = d[n*m-1]
        return distance

    @classmethod
    def minimum(cls, a: int, b: int, c: int) -> int:
        """
        Gets the minimum of three values

        Args:
            a (int):
            b (int):
            c (int):
        Returns:
            int: Minimum of given value.
        """
        return min(a, b, c)




import unittest

from uvm.dpi.uvm_hdl import uvm_hdl

class MemSlot():
    def __init__(self):
        self.value = 0
    def __int__(self):
        return self.value

class MockMem():
    def __init__(self, size):
        self.size = size
        self.mem_sig = [MemSlot()] * size

class MockDut():
    def __init__(self, top=True):
        self.signal = MemSlot()
        self.mem = MockMem(16)
        self.mem_2d = [MockMem(8), MockMem(8)]
        if top is True:
            self.sub_dut = MockDut(top=False)


class TestUVMHDL(unittest.TestCase):
    pass

    #def test_uvm_hdl_split_path(self):
    #    path1 = "dut.mem"
    #    spl = uvm_hdl.split_path(path1)
    #    self.assertEqual(spl, ['dut', 'mem'])
    #    path2 = "dut.mem[3]"
    #    spl = uvm_hdl.split_path(path2)
    #    self.assertEqual(spl, ['dut', 'mem', '[3]'])
    #    path3 = "dut.mem[3][7]"
    #    spl = uvm_hdl.split_path(path3)
    #    self.assertEqual(spl, ['dut', 'mem', '[3]', '[7]'])
    #    path4 = "dut.mem[5].ram_sig[9]"
    #    spl = uvm_hdl.split_path(path4)
    #    self.assertEqual(spl, ['dut', 'mem', '[5]', 'ram_sig', '[9]'])

    #def test_uvm_hdl_read(self):
    #    dut = MockDut()
    #    dut.mem.mem_sig[2] = 0x2
    #    val = []
    #    uvm_hdl.set_dut(dut)
    #    uvm_hdl.uvm_hdl_read('dut.mem.mem_sig[2]', val)
    #    self.assertEqual(val[0], 0x2)
    #    val = []
    #    dut.mem_2d[0].mem_sig[7] = 0xF
    #    uvm_hdl.uvm_hdl_read('dut.mem_2d[0].mem_sig[7]', val)
    #    self.assertEqual(val[0], 0xF)

    #def test_uvm_hdl_deposit(self):
    #    dut = MockDut()
    #    uvm_hdl.set_dut(dut)
    #    # Should raise an error
    #    # uvm_hdl.uvm_hdl_deposit('dut.error.signal', 0xABCD)
    #    uvm_hdl.uvm_hdl_deposit('dut.signal', 0xABCD)
    #    self.assertEqual(dut.signal.value, 0xABCD)

    #    uvm_hdl.uvm_hdl_deposit('dut.sub_dut.signal', 0xAFFF)
    #    self.assertEqual(dut.sub_dut.signal.value, 0xAFFF)

    #    for i in range(16):
    #        value = 0xF + 2 * i
    #        uvm_hdl.uvm_hdl_deposit('dut.mem.mem_sig[{}]'.format(i), value)
    #        msg = "uvm_hdl_deposit to mem{} OK".format(i)
    #        self.assertEqual(dut.mem.mem_sig[i].value, value, msg)



if __name__ == '__main__':
    unittest.main()



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import RisingEdge, Timer

from uvm.base import *
from uvm.comps import UVMDriver
from uvm.macros import uvm_component_utils, uvm_info
from ubus_transfer import *

#//------------------------------------------------------------------------------
#//
#// CLASS: ubus_master_driver
#//
#//------------------------------------------------------------------------------

def _print(msg):
    uvm_info("EEE_ubus_master_drv", msg, UVM_LOW)

class ubus_master_driver(UVMDriver):
    #
    #  // The virtual interface used to drive and view HDL signals.
    #  protected virtual ubus_if vifg
    #
    #  // Master Id
    #  protected int master_idg
    #

    #  // Provide implmentations of virtual methods such as get_type_name and create
    #  `uvm_component_utils_begin(ubus_master_driver)
    #    `uvm_field_int(master_id, UVM_DEFAULT)
    #  `uvm_component_utils_end

    def __init__(self, name, parent):
        UVMDriver.__init__(self, name, parent)
        # The virtual interface used to drive and view HDL signals.
        self.vif = None
        self.master_id = 0


    def build_phase(self, phase):
        super().build_phase(phase)
        arr = []
        if UVMConfigDb.get(self, "", "vif", arr):
            self.vif = arr[0]
        if self.vif is None:
            self.uvm_report_fatal("NOVIF", "virtual interface must be set for: " +
                self.get_full_name() + ".vif")
        arr = []
        if UVMConfigDb.get(self, "", "master_id", arr):
            self.master_id = arr[0]


    
    async def run_phase(self, phase):
        #    fork
        fork_get = cocotb.fork(self.get_and_drive())
        fork_reset = cocotb.fork(self.reset_signals())
        #await [fork_get, fork_reset.join()]
        await sv.fork_join([fork_get, fork_reset])
        #    join


    #  // get_and_drive
    #  virtual protected task get_and_drive()g
    
    async def get_and_drive(self):
        _print(" get_and_drive started")
        # @(negedge vif.sig_reset)g
        await RisingEdge(self.vif.sig_reset)
        _print(" got posedge reset")
        while True:
            await RisingEdge(self.vif.sig_clock)
            _print(" got rising edge of sig_clock")
            req = []
            await self.seq_item_port.get_next_item(req)
            _print(" got item from seq port: " + req[0].convert2string())
            #$cast(rsp, req.clone())g
            rsp = req[0].clone()
            rsp.set_id_info(req[0])
            await self.drive_transfer(rsp)
            self.seq_item_port.item_done()
            self.seq_item_port.put_response(rsp)
        #  endtask : get_and_drive


    #  // reset_signals
    
    async def reset_signals(self):
        _id = str(self.master_id)
        while True:
            await RisingEdge(self.vif.sig_reset)
            self.vif.rw = 0
            self.vif.sig_req[self.master_id] <= 0
            self.vif.sig_addr           <= 0
            self.vif.sig_data_out       <= 0
            self.vif.sig_size           <= 0
            self.vif.sig_read           <= 0
            self.vif.sig_write          <= 0
            self.vif.sig_bip            <= 0

    async def drive_transfer(self, trans):
        if trans.transmit_delay > 0:
            for i in range(trans.transmit_delay):
                await RisingEdge(self.vif.sig_clock)
        _print("Drivin trans into DUT now: " + trans.convert2string())
        await self.arbitrate_for_bus()
        await self.drive_address_phase(trans)
        await self.drive_data_phase(trans)


    def get_req_val(self, val):
        return int(self.vif.sig_req[self.master_id])
        #req_val = self.vif.sig_request.value.get_value()
        #if val == 1:
        #    req_val = req_val | (1 << self.master_id)
        #else:
        #    print("get_req_val read val " + str(req_val))
        #    req_val = req_val & ~(1 << self.master_id) & 0xFFFF
        #return req_val

    #  // arbitrate_for_bus
    
    async def arbitrate_for_bus(self):
        #self.vif.sig_request <= self.get_req_val(1)
        self.vif.sig_req[self.master_id] <= 1
        while True:
            await RisingEdge(self.vif.sig_clock)
            #grant_val = self.vif.sig_grant.value[self.master_id]
            grant_val = int(self.vif.sig_gnt[self.master_id])
            if grant_val != 0:
                break
        self.vif.sig_req[self.master_id] <= 0
        #self.vif.sig_request <= self.get_req_val(0)
        #  endtask : arbitrate_for_bus

    def is_grant_ok(self, grant_val):
        idx = self.master_id
        mask = 1 << idx
        result = mask & grant_val
        return result != 0

    #
    #  // drive_address_phase
    
    async def drive_address_phase(self, trans):
        self.vif.sig_addr <= trans.addr
        self.drive_size(trans.size)
        self.drive_read_write(trans.read_write)
        await RisingEdge(self.vif.sig_clock)
        #self.vif.sig_addr <= 32'bzg
        self.vif.sig_addr <= 0
        self.vif.sig_size <= 0
        self.vif.sig_read <= 0
        self.vif.sig_write <= 0
        #  endtask : drive_address_phase

    #
    #  // drive_data_phase
    
    async def drive_data_phase(self, trans):
        err = 0
        for i in range(trans.size):
            if (i == (trans.size - 1)):
                self.vif.sig_bip <= 0
            else:
                self.vif.sig_bip <= 1

            if trans.read_write == READ:
                arr = []
                await self.read_byte(arr, err)
                trans.data[i] = arr[0]
            else:
                await self.write_byte(trans.data[i], err)

        self.vif.sig_data_out <= 0
        self.vif.sig_bip <= 0
        #  endtask : drive_data_phase


    #  // read_byte
    async def read_byte(self, data, error):
        self.vif.rw <= 0
        while True:
            await RisingEdge(self.vif.sig_clock)
            if self.vif.sig_wait == 0:
                break
        uvm_info("READ_BYTE", "Got byte {}".format(hex(self.vif.sig_data.value)), UVM_LOW)
        data.append(self.vif.sig_data.value)
        #  endtask : read_byte


    #  // write_byte
    async def write_byte(self, data, error):
        uvm_info("MASTER_DRV", "Setting rw to 1 now", UVM_LOW)
        self.vif.rw <= 1
        self.vif.sig_data_out <= data
        while True:
            await RisingEdge(self.vif.sig_clock)
            if self.vif.sig_wait == 0:
                break
        self.vif.rw <= 0

    #  // drive_size
    def drive_size(self, size):
        if size == 1:
            self.vif.sig_size <= 0
        elif size == 2:
            self.vif.sig_size <= 1
        elif size == 4:
            self.vif.sig_size <= 2
        elif size == 8:
            self.vif.sig_size <= 3

    #  // drive_read_write
    def drive_read_write(self, rw):
        uvm_info("MASTER_DRV", "Driving rw-value: " + str(rw), UVM_LOW)

        if rw == NOP:
            self.vif.sig_read <= 0
            self.vif.sig_write <= 0
        elif rw == READ:
            self.vif.sig_read <= 1
            self.vif.sig_write <= 0
        elif rw == WRITE:
            self.vif.sig_read <= 0
            self.vif.sig_write <= 1


uvm_component_utils(ubus_master_driver)
